"""
Functionality relating to a generic remote library.
"""
import logging
from abc import ABCMeta, abstractmethod
from collections.abc import Collection, Mapping, Iterable
from typing import Any, Literal, Self

from musify.base import MusifyItem
from musify.libraries.core.object import Track, Library, Playlist
from musify.libraries.remote.core.api import RemoteAPI
from musify.libraries.remote.core.factory import RemoteObjectFactory
from musify.libraries.remote.core.object import RemoteCollection, SyncResultRemotePlaylist
from musify.libraries.remote.core.object import RemoteTrack, RemotePlaylist, RemoteArtist, RemoteAlbum
from musify.libraries.remote.core.types import RemoteObjectType
from musify.logger import MusifyLogger
from musify.logger import STAT
from musify.processors.base import Filter
from musify.processors.filter import FilterDefinedList
from musify.utils import align_string, get_max_width, to_collection

type RestorePlaylistsType = (
        Library |
        Collection[Playlist] |
        Mapping[str, Iterable[Track]] |
        Mapping[str, Iterable[str]] |
        Mapping[str, Iterable[Mapping[str, Any]]]
)
type SyncPlaylistsType = Library | Mapping[str, Collection[MusifyItem]] | Collection[Playlist] | None


class RemoteLibrary[
    A: RemoteAPI, PL: RemotePlaylist, TR: RemoteTrack, AL: RemoteAlbum, AR: RemoteArtist
](RemoteCollection[TR], Library[TR], metaclass=ABCMeta):
    """
    Represents a remote library, providing various methods for manipulating
    tracks and playlists across an entire remote library collection.

    :param api: The instantiated and authorised API object for this source type.
    :param playlist_filter: An optional :py:class:`Filter` to apply or collection of playlist names to include when
        loading playlists. Playlist names will be passed to this filter to limit which playlists are loaded.
    """

    __slots__ = ("logger", "_factory", "_playlists", "_tracks", "_albums", "_artists", "playlist_filter")
    __attributes_classes__ = (Library, RemoteCollection)
    __attributes_ignore__ = ("api", "factory")

    @property
    def _log_min_width(self) -> int:
        max_type_width = max(len(str(enum.name)) for enum in RemoteObjectType.all())
        return len(f"USER'S {self.api.source.upper()} ") + max_type_width

    @property
    def factory(self) -> RemoteObjectFactory[A, PL, TR, AL, AR]:
        """Stores the key object classes for a remote source."""
        return self._factory

    @abstractmethod
    def _create_factory(self, api: A) -> RemoteObjectFactory[A, PL, TR, AL, AR]:
        """Stores the key object classes for a remote source."""
        raise NotImplementedError

    @property
    def api(self):
        """Authorised API object for making authenticated calls to a user's library"""
        return self.factory.api

    @property
    def name(self):
        """The username associated with this library"""
        return self.api.user_name

    @property
    def id(self):
        """The user ID associated with this library"""
        return self.api.user_id

    @property
    def playlists(self) -> dict[str, PL]:
        return self._playlists

    @property
    def tracks(self) -> list[TR]:
        """All user's saved tracks"""
        return self._tracks

    @property
    def artists(self) -> list[AR]:
        """All user's saved artists"""
        return self._artists

    @property
    def albums(self) -> list[AL]:
        """All user's saved albums"""
        return self._albums

    def __init__(self, api: A, playlist_filter: Collection[str] | Filter[str] = ()):
        super().__init__()

        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

        if not isinstance(playlist_filter, Filter):
            playlist_filter = FilterDefinedList(playlist_filter)
        #: :py:class:`Filter` to filter out the playlists loaded by name.
        self.playlist_filter: Filter[str] = playlist_filter

        self._factory = self._create_factory(api=api)
        self._playlists: dict[str, PL] = {}
        self._tracks: list[TR] = []
        self._albums: list[AL] = []
        self._artists: list[AR] = []

    async def __aenter__(self) -> Self:
        await self.api.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.api.__aexit__(exc_type, exc_val, exc_tb)

    async def extend(self, __items: Iterable[MusifyItem], allow_duplicates: bool = True) -> None:
        self.logger.debug(f"Extend {self.api.source} tracks data: START")

        load_uris = []
        for item in __items:
            if not allow_duplicates and item in self.items:
                continue
            elif isinstance(item, RemoteTrack):
                self.items.append(item)
            elif item.has_uri:
                load_uris.append(item.uri)

        if not load_uris:
            return

        self.logger.info(
            f"\33[1;95m  >\33[1;97m Extending {self.api.source} library "
            f"with {len(load_uris)} additional tracks \33[0m"
        )

        load_tracks = await self.api.get_tracks(load_uris, features=True)
        self.items.extend(map(self.factory.track, load_tracks))

        self.logger.print_line(STAT)
        self.log_tracks()
        self.logger.print_line()
        self.logger.debug(f"Extend {self.api.source} tracks data: DONE\n")

    async def load(self) -> None:
        """Loads all data from the remote API for this library and log results."""
        self.logger.debug(f"Load {self.api.source} library: START")
        self.logger.info(f"\33[1;95m ->\33[1;97m Loading {self.api.source} library \33[0m")

        await self.load_playlists()
        await self.load_tracks()
        await self.load_saved_albums()
        await self.load_saved_artists()

        self.logger.print_line(STAT)
        self.log_playlists()
        self.log_tracks()
        self.log_albums()
        self.log_artists()

        self.logger.print_line()
        self.logger.debug(f"Load {self.api.source} library: DONE\n")

    ###########################################################################
    ## Load - playlists
    ###########################################################################
    async def load_playlists(self) -> None:
        """
        Load all playlists from the API that match the filter rules in this library. Also loads all their tracks.
        WARNING: Overwrites any currently loaded playlists.
        """
        self.logger.debug(f"Load {self.api.source} playlists: START")

        responses = await self.api.get_user_items(kind=RemoteObjectType.PLAYLIST)
        responses = self._filter_playlists(responses)

        self.logger.info(
            f"\33[1;95m  >\33[1;97m Getting {self._get_total_tracks(responses=responses)} "
            f"tracks from {len(responses)} {self.api.source} playlists \33[0m"
        )
        await self.api.get_items(responses, kind=RemoteObjectType.PLAYLIST)

        playlists = [
            self.factory.playlist(response=r, skip_checks=False)
            for r in self.logger.get_synchronous_iterator(responses, desc="Processing playlists", unit="playlists")
        ]

        self._playlists = {pl.name: pl for pl in sorted(playlists, key=lambda x: x.name.casefold())}
        self.logger.debug(f"Load {self.api.source} playlists: DONE")

    def _filter_playlists(self, responses: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """
        Filter API responses from all playlists according to filter rules in this library.

        :return: Filtered API responses.
        """
        raise NotImplementedError

    def _get_total_tracks(self, responses: list[dict[str, Any]]) -> int:
        """
        Returns the total number of tracks across all given playlists data. Used for logging purposes

        :return: Total track count.
        """
        raise NotImplementedError

    def log_playlists(self) -> None:
        max_width = get_max_width(self.playlists, min_width=self._log_min_width)

        self.logger.stat(f"\33[1;96m{self.api.source.upper()} PLAYLISTS: \33[0m")
        for name, playlist in self.playlists.items():
            name = align_string(playlist.name, max_width=max_width)
            self.logger.stat(f"\33[97m{name} \33[0m| \33[92m{len(playlist):>6} total tracks \33[0m")

    ###########################################################################
    ## Load - tracks
    ###########################################################################
    async def load_tracks(self) -> None:
        """
        Load all user's saved tracks from the API.
        Updates currently loaded tracks in-place or appends if not already loaded.
        """
        self.logger.debug(f"Load user's saved {self.api.source} tracks: START")

        responses = await self.api.get_user_items(kind=RemoteObjectType.TRACK)
        for response in self.logger.get_synchronous_iterator(responses, desc="Processing tracks", unit="tracks"):
            track = self.factory.track(response=response, skip_checks=True)

            if not track.has_uri:  # skip any invalid non-remote responses
                continue

            current = next((item for item in self._tracks if item == track), None)
            if current is None:
                self._tracks.append(track)
                continue

            current._response = track.response
            current.refresh(skip_checks=False)

        self.logger.debug(f"Load user's saved {self.api.source} tracks: DONE")

    async def enrich_tracks(self, *_, **__) -> None:
        """
        Call API to enrich elements of track objects improving metadata coverage.
        This is an optionally implementable method. Defaults to doing nothing.
        """
        self.logger.debug("Enrich tracks not implemented for this library, skipping...")

    def log_tracks(self) -> None:
        playlist_tracks = [track.uri for tracks in self.playlists.values() for track in tracks]
        in_playlists = sum(1 for track in self.tracks if track.uri in playlist_tracks)
        album_tracks = [track.uri for tracks in self.albums for track in tracks]
        in_albums = sum(1 for track in self.tracks if track.uri in album_tracks)

        width = get_max_width(self.playlists, min_width=self._log_min_width)
        self.logger.stat(
            f"\33[1;96m{"USER'S " + self.api.source.upper() + " TRACKS":<{width}}\33[1;0m |"
            f"\33[92m{in_playlists:>7} in playlists  \33[0m|"
            f"\33[92m{in_albums:>7} in saved albums \33[0m|"
            f"\33[1;94m{len(self.tracks):>7} total tracks \33[0m"
        )

    ###########################################################################
    ## Load - albums
    ###########################################################################
    async def load_saved_albums(self) -> None:
        """
        Load all user's saved albums from the API.
        Updates currently loaded albums in-place or appends if not already loaded.
        """
        self.logger.debug(f"Load user's saved {self.api.source} albums: START")

        responses = await self.api.get_user_items(kind=RemoteObjectType.ALBUM)
        for response in self.logger.get_synchronous_iterator(responses, desc="Processing albums", unit="albums"):
            album = self.factory.album(response=response, skip_checks=True)

            current = next((item for item in self._albums if item == album), None)
            if current is None:
                self._albums.append(album)
            else:
                current._response = album.response
                current.refresh(skip_checks=True)

            for track in album.tracks:  # add tracks from this album to the user's saved tracks
                if track not in self.tracks:
                    self._tracks.append(track)

        self.logger.debug(f"Load user's saved {self.api.source} albums: DONE")

    async def enrich_saved_albums(self, *_, **__) -> None:
        """
        Call API to enrich elements of user's saved album objects improving metadata coverage.
        This is an optionally implementable method. Defaults to doing nothing.
        """
        self.logger.debug("Enrich albums not implemented for this library, skipping...")

    def log_albums(self) -> None:
        """Log stats on currently loaded albums"""
        width = get_max_width(self.playlists, min_width=self._log_min_width)
        self.logger.stat(
            f"\33[1;96m{"USER'S " + self.api.source.upper() + " ALBUMS":<{width}}\33[1;0m |"
            f"\33[92m{sum(len(album.tracks) for album in self.albums):>7} album tracks  \33[0m|"
            f"\33[92m{sum(len(album.artists) for album in self.albums):>7} album artists   \33[0m|"
            f"\33[1;94m{len(self.artists):>7} total albums \33[0m"
        )

    ###########################################################################
    ## Load - artists
    ###########################################################################
    async def load_saved_artists(self) -> None:
        """
        Load all user's saved artists from the API.
        Updates currently loaded artists in-place or appends if not already loaded.
        """
        self.logger.debug(f"Load user's saved {self.api.source} artists: START")

        responses = await self.api.get_user_items(kind=RemoteObjectType.ARTIST)
        for response in self.logger.get_synchronous_iterator(responses, desc="Processing artists", unit="artists"):
            artist = self.factory.artist(response=response, skip_checks=True)

            current = next((item for item in self._artists if item == artist), None)
            if current is None:
                self._artists.append(artist)
                continue

            current._response = artist.response
            current.refresh(skip_checks=True)

        self.logger.debug(f"Load user's saved {self.api.source} artists: DONE")

    async def enrich_saved_artists(self, *_, **__) -> None:
        """
        Call API to enrich elements of user's saved artist objects improving metadata coverage.
        This is an optionally implementable method. Defaults to doing nothing.
        """
        self.logger.debug("Enrich artists not implemented for this library, skipping...")

    def log_artists(self) -> None:
        """Log stats on currently loaded artists"""
        width = get_max_width(self.playlists, min_width=self._log_min_width)
        self.logger.stat(
            f"\33[1;96m{"USER'S " + self.api.source.upper() + " ARTISTS":<{width}}\33[1;0m |"
            f"\33[92m{sum(len(artist.tracks) for artist in self.artists):>7} artist tracks \33[0m|"
            f"\33[92m{sum(len(artist.albums) for artist in self.artists):>7} artist albums   \33[0m|"
            f"\33[1;94m{len(self.artists):>7} total artists \33[0m"
        )

    ###########################################################################
    ## Backup/Restore/Sync
    ###########################################################################
    def backup_playlists(self) -> dict[str, list[str]]:
        """
        Produce a backup map of <playlist name>: [<URIs] for all playlists in this library
        which can be saved to JSON for backup purposes.
        """
        return {name: [track.uri for track in pl] for name, pl in self.playlists.items()}

    async def restore_playlists(self, playlists: RestorePlaylistsType, dry_run: bool = True) -> None:
        """
        Restore playlists from a backup to loaded playlist objects.

        This function does not sync the updated playlists with the remote library.

        When ``dry_run`` is False, this function does create new playlists on the remote library for playlists
        given that do not exist in this Library.

        :param playlists: Values that represent the playlists to restore from.
        :param dry_run: When True, do not create playlists
            and just skip any playlists that are not already currently loaded.
        """
        playlists = self._extract_playlists_from_backup(playlists)
        uri_tracks = {track.uri: track for track in self.tracks}
        uri_get = [uri for uri_list in playlists.values() for uri in uri_list if uri not in uri_tracks]

        if uri_get:
            tracks_data = await self.api.get_tracks(uri_get, features=False)
            tracks = list(map(self.factory.track, tracks_data))
            uri_tracks |= {track.uri: track for track in tracks}

        async def _get_playlist(n: str) -> PL | None:
            pl = self.playlists.get(n)
            if not pl and dry_run:  # skip on dry run
                return
            if not pl:  # new playlist given, create it on remote first
                # noinspection PyArgumentList
                pl = await self.factory.playlist.create(name=n)

            return pl

        bar = self.logger.get_asynchronous_iterator(map(_get_playlist, playlists), disable=True)
        playlists_remote: dict[str, PL] = {pl.name: pl for pl in await bar if pl is not None}

        for name, uri_list in playlists.items():
            playlist = playlists_remote.get(name)
            if playlist is None:
                continue

            playlist._tracks = [uri_tracks.get(uri) for uri in uri_list]
            self.playlists[name] = playlist

    @staticmethod
    def _extract_playlists_from_backup(playlists: RestorePlaylistsType) -> Mapping[str, list[str]]:
        if isinstance(playlists, Library):  # get URIs from playlists in library
            playlists = {name: [track.uri for track in pl] for name, pl in playlists.playlists.items()}
        elif (
                isinstance(playlists, Mapping)
                and all(isinstance(v, MusifyItem) for vals in playlists.values() for v in vals)
        ):
            # get URIs from playlists in map values
            playlists = {name: [item.uri for item in pl] for name, pl in playlists.items()}
        elif isinstance(playlists, Mapping):
            # get URIs from playlists in collection
            playlists = {
                name:
                    [t["uri"] if isinstance(t, Mapping) else t for t in tracks["tracks"]]
                    if isinstance(tracks, Mapping) else tracks
                for name, tracks in playlists.items()
            }
        elif isinstance(playlists, Collection):
            # get URIs from playlists in collection
            playlists = {pl.name: [track.uri for track in pl] for pl in playlists}
        else:
            playlists = {name: to_collection(tracks, list) for name, tracks in playlists.items()}

        return playlists

    async def sync(
            self,
            playlists: SyncPlaylistsType = None,
            kind: Literal["new", "refresh", "sync"] = "new",
            reload: bool = True,
            dry_run: bool = True
    ) -> dict[str, SyncResultRemotePlaylist]:
        """
        Synchronise this playlist object with the remote playlist it is associated with.

        Clear options:
            * 'new': Do not clear any items from the remote playlist and only add any tracks
                from this playlist object not currently in the remote playlist.
            * 'refresh': Clear all items from the remote playlist first, then add all items from this playlist object.
            * 'sync': Clear all items not currently in this object's items list, then add all tracks
                from this playlist object not currently in the remote playlist.

        :param playlists: Provide a library, map of playlist name to playlist or collection of playlists
            to synchronise to the remote library.
            Use the currently loaded ``playlists`` in this object if not given.
        :param kind: Sync option for the remote playlist. See description.
        :param reload: When True, once synchronisation is complete, reload this RemotePlaylist object
            to reflect the changes on the remote playlist if enabled. Skip if False.
        :param dry_run: Run function, but do not modify the remote playlists at all.
        :return: Map of playlist name to the results of the sync as a :py:class:`SyncResultRemotePlaylist` object.
        """
        self.logger.debug(f"Sync {self.api.source} playlists: START")

        playlists = self._extract_playlists_for_sync(playlists)

        log_kind = "adding new items only"
        if kind != "new":
            log_kind = 'all' if kind == 'refresh' else 'extra'
            log_kind = f"clearing {log_kind} items from each {self.api.source} playlist first"
        self.logger.info(
            f"\33[1;95m ->\33[1;97m Synchronising {len(playlists)} {self.api.source} playlists: {log_kind}"
            f"{f" and reloading stored playlists" if reload else ""} \33[0m"
        )

        async def _sync_playlist(name: str, pl: Collection[MusifyItem]) -> tuple[str, SyncResultRemotePlaylist]:
            if name not in self.playlists:  # new playlist given, create it on remote first
                if dry_run:
                    result = SyncResultRemotePlaylist(
                        start=0, added=len(pl), removed=0, unchanged=0, difference=len(pl), final=len(pl)
                    )
                    return name, result

                # noinspection PyArgumentList
                self.playlists[name] = await self.factory.playlist.create(name=name)

            return name, await self.playlists[name].sync(items=pl, kind=kind, reload=reload, dry_run=dry_run)

        bar = self.logger.get_asynchronous_iterator(
            [_sync_playlist(name, pl) for name, pl in playlists.items()],
            desc=f"Synchronising {self.api.source}",
            unit="playlists"
        )
        results = dict(await bar)

        self.logger.print_line()
        self.logger.debug(f"Sync {self.api.source} playlists: DONE\n")
        return results

    def _extract_playlists_for_sync(self, playlists: SyncPlaylistsType) -> Mapping[str, Collection[MusifyItem]]:
        if not playlists:  # use the playlists as stored in this library object
            playlists = self.playlists
        elif isinstance(playlists, Library):  # get map of playlists from the given library
            playlists = playlists.playlists
        elif isinstance(playlists, Collection) and all(isinstance(pl, Playlist) for pl in playlists):
            # reformat list to map
            playlists = {pl.name: pl for pl in playlists}

        return playlists

    def log_sync(self, results: SyncResultRemotePlaylist | Mapping[str, SyncResultRemotePlaylist]) -> None:
        """Log stats from the results of a ``sync`` operation"""
        if not results:
            return
        if not isinstance(results, Mapping):
            results = {"": results}

        max_width = get_max_width(results)

        self.logger.stat(f"\33[1;96mSync {self.api.source} playlists' stats: \33[0m")
        for name, result in results.items():
            self.logger.stat(
                f"\33[97m{align_string(name, max_width=max_width)} \33[0m|"
                f"\33[96m{result.start:>6} initial \33[0m|"
                f"\33[92m{result.added:>6} added \33[0m|"
                f"\33[91m{result.removed:>6} removed \33[0m|"
                f"\33[93m{result.unchanged:>6} unchanged \33[0m|"
                f"\33[94m{result.difference:>6} difference \33[0m|"
                f"\33[1;97m{result.final:>6} final \33[0m"
            )



"""
Generic base classes and functions for file operations.
"""
from abc import ABCMeta, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Any

from musify.file.exception import InvalidFileType, FileDoesNotExistError


class File(metaclass=ABCMeta):
    """Generic class for representing a file on a system."""

    __slots__ = ()

    #: Extensions of files that can be loaded by this class.
    valid_extensions: frozenset[str]

    @property
    @abstractmethod
    def path(self) -> Path:
        """The path to the file."""
        raise NotImplementedError

    @property
    def folder(self) -> str:
        """The parent folder of the file."""
        return self.path.parent.name

    @property
    def filename(self) -> str:
        """The filename without extension."""
        return self.path.stem

    @property
    def ext(self) -> str:
        """The file extension in lowercase."""
        return self.path.suffix

    @property
    def size(self) -> int | None:
        """The size of the file in bytes"""
        return self.path.stat().st_size if self.path.is_file() else None

    @property
    def date_created(self) -> datetime | None:
        """:py:class:`datetime` object representing when the file was created"""
        return datetime.fromtimestamp(self.path.stat().st_ctime) if self.path.is_file() else None

    @property
    def date_modified(self) -> datetime | None:
        """:py:class:`datetime` object representing when the file was last modified"""
        return datetime.fromtimestamp(self.path.stat().st_mtime) if self.path.is_file() else None

    @classmethod
    def _validate_type(cls, path: Path) -> None:
        """Raises an exception if the ``path`` extension is not accepted"""
        if path.suffix not in cls.valid_extensions:
            raise InvalidFileType(
                path.suffix,
                f"Not an accepted {cls.__name__} file extension. "
                f"Use only: {', '.join(cls.valid_extensions)}"
            )

    @staticmethod
    def _validate_existence(path: Path):
        """Raises an exception if there is no file at the given ``path``"""
        if not path.is_file():
            raise FileDoesNotExistError(path)

    @classmethod
    def get_filepaths(cls, folder: str | Path) -> set[Path]:
        """Get all files in a given folder that match this File object's valid filetypes recursively."""
        paths: set[Path] = set()
        folder = Path(folder)

        for ext in cls.valid_extensions:
            paths |= set(folder.rglob(str(Path("**", f"[!.]*{ext}"))))
            # paths |= set(folder.rglob(str(Path("**", f".*{ext}"))))

        # do not return paths in the recycle bin in Windows-based folders
        return {path for path in paths if "$RECYCLE.BIN" not in path.parts}

    @abstractmethod
    async def load(self, *args, **kwargs) -> Any:
        """Load the file to this object"""
        raise NotImplementedError

    @abstractmethod
    async def save(self, dry_run: bool = True, *args, **kwargs) -> Any:
        """
        Save this object to file.

        :param dry_run: Run function, but do not modify the file on the disk.
        """
        raise NotImplementedError

    def __hash__(self):
        """Uniqueness of a file is its path"""
        return hash(self.path)



"""
Generic utility functions and classes which can be used throughout the entire package.
"""
import re
from collections import Counter
from collections.abc import Iterable, Collection, MutableSequence, Mapping, MutableMapping
from typing import Any

import unicodedata

from musify.exception import MusifyTypeError, MusifyImportError
from musify.types import Number


###########################################################################
## Extended primitives
###########################################################################
class SafeDict(dict):
    """Extends dict to ignore missing keys when using format_map operations"""

    __slots__ = ()

    def __missing__(self, key):
        return "{" + key + "}"


###########################################################################
## String
###########################################################################
def strip_ignore_words(value: str, words: Iterable[str] | None = frozenset({"The", "A"})) -> tuple[bool, str]:
    """
    Remove ignorable words from the beginning of a string.

    Useful for sorting collections strings with ignorable start words and/or special characters.
    Only removes the first word it finds at the start of the string.

    :return: Tuple of (True if the string starts with some special character, the formatted string)
    """
    if not value:
        return False, value

    special_chars = list('!"£$%^&*()_+-=…')
    special_start = any(value.startswith(c) for c in special_chars)
    value = re.sub(r"^\W+", "", value).strip()

    if not words:
        return not special_start, value

    new_value = value
    for word in words:
        new_value = re.sub(rf"^{word}\s+", "", value, flags=re.I)
        if new_value != value:
            break

    return not special_start, new_value


def safe_format_map[T](value: T, format_map: Mapping[str, Any]) -> T:
    """
    Apply a ``format_map`` to a given ``value`` ignoring missing keys.
    If ``value`` is a map, apply the ``format_map`` recursively.
    """
    if not isinstance(format_map, SafeDict):
        format_map = SafeDict(**format_map)

    if isinstance(value, MutableMapping):
        for k, v in value.items():
            value[k] = safe_format_map(v, format_map)
    elif isinstance(value, str) and '{' in value and '}' in value:
        value = value.format_map(format_map)
    return value


_UNICODE_MODIFIERS = {"Mc", "Mn"}
_UNICODE_DOUBLE_WIDTH = {"So"}


def unicode_len(value: str) -> int:
    """
    Returns the visible length of ``value`` when rendered with fixed-width font.

    Takes into account unicode characters which render with varying widths.
    """
    length = sum(unicodedata.category(char) not in _UNICODE_MODIFIERS for char in value)
    length += sum(unicodedata.category(char) in _UNICODE_DOUBLE_WIDTH for char in value)  # takes up 2 spaces
    return length


def unicode_reversed(value: str) -> str:
    """Returns a reversed string or ``value`` keeping unicode characters which combine in the correct order"""
    value_reversed = ""
    char_combined = ""

    for i, char in enumerate(value):
        char_combined += char
        if i + 1 < len(value) and unicodedata.category(value[i + 1]) in _UNICODE_MODIFIERS:
            continue

        value_reversed = char_combined + value_reversed
        char_combined = ""
    return value_reversed


def get_max_width(values: Collection[Any], min_width: int = 15, max_width: int = 50) -> int:
    """
    Get max width of given list of ``values`` for column-aligned logging.

    Uses width as would be seen in a fixed-width font taking into account characters with varying widths.
    """
    if len(values) == 0:
        return min_width
    max_len = unicode_len(max(map(str, values), key=unicode_len))
    return limit_value(value=max_len + 1, floor=min_width, ceil=max_width)


def align_string(value: Any, max_width: int = 0, truncate_left: bool = False) -> str:
    """
    Align string with space padding and truncate any string longer than ``max_width`` with ``...``

    This function aligns based on fixed-width fonts.
    Therefore, unicode characters (e.g. emojis) will be aligned based on their width in a fixed-width font.

    :param value: The value to be aligned. Will first be converted to a string.
    :param max_width: The expected width (i.e. number of fixed-width characters)
        the string should occupy in a fixed-width font.
    :param truncate_left: When truncating, truncate the left (i.e. start) of the string.
    :return: The padded and truncated string with visible length == ``max_width``.
    """
    value_str = str(value)
    if not value_str or max_width == 0:
        return " " * max_width

    if truncate_left:  # reverse string for truncate right operations
        value_str = unicode_reversed(value_str)

    value_truncated = ""
    dots_count = limit_value(max_width - 3, 0, 3) if max_width < unicode_len(value_str) else 0
    expected_len = max_width - dots_count
    for char in value_str:
        # stop on double width characters and extend dots_count to cover missing character if needed
        if unicodedata.category(char) in _UNICODE_DOUBLE_WIDTH and unicode_len(value_truncated + char) > expected_len:
            dots_count += 1 if dots_count < 3 and unicode_len(value_truncated) < max_width else 0
            break

        # always add unicode modifiers even if unicode_len == expected_width
        if unicodedata.category(char) not in _UNICODE_MODIFIERS and unicode_len(value_truncated) == expected_len:
            break

        value_truncated += char

    value_truncated += "." * dots_count  # add ellipses

    if truncate_left:  # reverse back
        value_truncated = unicode_reversed(value_truncated)

    # extend max_width with difference in length from actual len to unicode_len
    max_width += len(value_truncated) - unicode_len(value_truncated)
    return f"{value_truncated:<{max_width}.{max_width}}"


###########################################################################
## Number
###########################################################################
def limit_value(value: Number, floor: Number = 1, ceil: Number = 50) -> Number:
    """Limit a given ``value`` to always be between some ``floor`` and ``ceil``"""
    return max(min(value, ceil), floor)


###########################################################################
## Collection
###########################################################################
def to_collection[T: (list, set, tuple)](data: Any, cls: type[T] = tuple) -> T | None:
    """
    Safely turn any object into a collection of a given type ``T``.

    Strings are converted to collections of size 1 where the first element is the string.
    Returns None if value is None.
    """
    if data is None or isinstance(data, cls):
        return data
    elif isinstance(data, Iterable) and not isinstance(data, str) and not isinstance(data, Mapping):
        return cls(data)
    elif cls is tuple:
        return (data,)
    elif cls is set:
        return {data}
    elif cls is list:
        return [data]
    raise MusifyTypeError(f"Unable to convert data to {cls.__name__} (data={data})")


###########################################################################
## Mapping
###########################################################################
def flatten_nested[T: Any](nested: MutableMapping, previous: MutableSequence[T] | None = None) -> list[T]:
    """Flatten the final layers of the values of a nested map to a single list"""
    if previous is None:
        previous = []

    if isinstance(nested, MutableMapping):
        for key, value in nested.items():
            flatten_nested(value, previous=previous)
    elif isinstance(nested, (list, set, tuple)):
        previous.extend(nested)
    else:
        previous.append(nested)

    return previous


def merge_maps[T: MutableMapping](source: T, new: Mapping, extend: bool = True, overwrite: bool = False) -> T:
    """
    Recursively update a given ``source`` map in place with a ``new`` map.

    :param source: The source map.
    :param new: The new map with values to update for the source map.
    :param extend: When a value is a list and a list is already present in the source map, extend the list when True.
        When False, only replace the list if overwrite is True.
    :param overwrite: When True, overwrite any value in the source list destructively.
    :return: The updated dict.
    """
    def is_collection(value: Any) -> bool:
        """Return True if ``value`` is of type ``Collection`` and not a string or map"""
        return isinstance(value, Collection) and not isinstance(value, str) and not isinstance(value, Mapping)

    for k, v in new.items():
        if isinstance(v, Mapping) and isinstance(source.get(k, {}), Mapping):
            source[k] = merge_maps(source.get(k, {}), v, extend=extend, overwrite=overwrite)
        elif extend and is_collection(v) and is_collection(source.get(k, [])):
            source[k] = to_collection(source.get(k, []), list) + to_collection(v, list)
        elif overwrite or source.get(k) is None:
            source[k] = v
    return source


###########################################################################
## Misc
###########################################################################
def get_most_common_values(values: Iterable[Any]) -> list[Any]:
    """Get an ordered list of the most common values for a given collection of ``values``"""
    return [x[0] for x in Counter(values).most_common()]


def get_user_input(text: str | None = None) -> str:
    """Print formatted dialog with optional text and get the user's input."""
    return input(f"\33[93m{text}\33[0m | ").strip()


def required_modules_installed(modules: list, this: object = None) -> bool:
    """Check the required modules are installed, raise :py:class:`MusifyImportError` if not."""
    modules_installed = all(module is not None for module in modules)
    if not modules_installed and this is not None:
        names = [name for name, obj in globals().items() if obj in modules and not name.startswith("_")]
        if isinstance(this, str):
            message = f"Cannot run {this}. Required modules: {", ".join(names)}"
        else:
            message = f"Cannot create {this.__class__.__name__} object. Required modules: {", ".join(names)}"

        raise MusifyImportError(message)

    return modules_installed



"""
Processor operations that search for and match given items with remote items.

Searches for matches on remote APIs, matches the item to the best matching result from the query,
and assigns the ID of the matched object back to the item.
"""
import logging
from collections.abc import Mapping, Sequence, Iterable, Collection, Awaitable
from dataclasses import dataclass, field
from typing import Any, Self

from musify.base import MusifyObject, MusifyItemSettable, Result
from musify.exception import MusifyAttributeError
from musify.field import TagField, TagFields as Tag
from musify.libraries.core.collection import MusifyCollection
from musify.libraries.remote.core.api import RemoteAPI
from musify.libraries.remote.core.factory import RemoteObjectFactory
from musify.libraries.remote.core.types import RemoteObjectType
from musify.logger import MusifyLogger
from musify.logger import REPORT
from musify.processors.base import Processor
from musify.processors.match import ItemMatcher
from musify.types import UnitIterable
from musify.utils import align_string, get_max_width


@dataclass(frozen=True)
class ItemSearchResult[T: MusifyItemSettable](Result):
    """Stores the results of the searching process."""
    #: Sequence of Items for which matches were found from the search.
    matched: Sequence[T] = field(default=tuple())
    #: Sequence of Items for which matches were not found from the search.
    unmatched: Sequence[T] = field(default=tuple())
    #: Sequence of Items which were skipped during the search.
    skipped: Sequence[T] = field(default=tuple())


@dataclass(frozen=True)
class SearchConfig:
    """Key settings related to a search algorithm."""
    #: The fields to match results on.
    match_fields: TagField | Iterable[TagField]
    #: A sequence of the tag names to use as search fields in the 1st pass.
    search_fields_1: Sequence[TagField] = (Tag.NAME,)
    #: If no results are found from the tag names in ``search_fields_1`` on the 1st pass,
    #: an optional sequence of the tag names to use as search fields in the 2nd pass.
    search_fields_2: Iterable[TagField] = ()
    #: If no results are found from the tag names in ``search_fields_2`` on the 2nd pass,
    #: an optional sequence of the tag names to use as search fields in the 3rd pass.
    search_fields_3: Iterable[TagField] = ()

    #: The number of the results to request when querying the API.
    result_count: int = 10
    #: The minimum acceptable score for an item to be considered a match.
    min_score: float = 0.1
    #: The maximum score for an item to be considered a perfect match.
    #: After this score is reached by an item, any other items are disregarded as potential matches.
    max_score: float = 0.8
    #: When True, items determined to be karaoke are allowed when matching added items.
    #: Skip karaoke results otherwise.
    allow_karaoke: bool = False


class RemoteItemSearcher(Processor):
    """
    Searches for remote matches for a list of item collections.

    :param matcher: The :py:class:`ItemMatcher` to use when comparing any changes made by the user in remote playlists
        during the checking operation
    :param object_factory: The :py:class:`RemoteObjectFactory` to use when creating new remote objects.
        This must have a :py:class:`RemoteAPI` assigned for this processor to work as expected.
    """

    __slots__ = ("logger", "matcher", "factory")

    #: The :py:class:`SearchSettings` for each :py:class:`RemoteObjectType`
    search_settings: dict[RemoteObjectType, SearchConfig] = {
        RemoteObjectType.TRACK: SearchConfig(
            match_fields={Tag.TITLE, Tag.ARTIST, Tag.ALBUM, Tag.LENGTH},
            search_fields_1=[Tag.NAME, Tag.ARTIST],
            search_fields_2=[Tag.NAME, Tag.ALBUM],
            search_fields_3=[Tag.NAME],
            result_count=10,
            min_score=0.1,
            max_score=0.8,
            allow_karaoke=False,
        ),
        RemoteObjectType.ALBUM: SearchConfig(
            match_fields={Tag.ARTIST, Tag.ALBUM, Tag.LENGTH},
            search_fields_1=[Tag.NAME, Tag.ARTIST],
            search_fields_2=[Tag.NAME],
            result_count=5,
            min_score=0.1,
            max_score=0.7,
            allow_karaoke=False,
        )
    }

    @property
    def api(self) -> RemoteAPI:
        """The :py:class:`RemoteAPI` to call"""
        return self.factory.api

    def __init__(self, matcher: ItemMatcher, object_factory: RemoteObjectFactory):
        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

        #: The :py:class:`ItemMatcher` to use when comparing any changes made by the user in remote playlists
        #: during the checking operation
        self.matcher = matcher
        #: The :py:class:`RemoteObjectFactory` to use when creating new remote objects.
        self.factory = object_factory

    async def __aenter__(self) -> Self:
        await self.api.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.api.__aexit__(exc_type, exc_val, exc_tb)

    async def _get_results(
            self, item: MusifyObject, kind: RemoteObjectType, settings: SearchConfig
    ) -> list[dict[str, Any]] | None:
        """Query the API to get results for the current item based on algorithm settings"""
        self.matcher.clean_tags(item)

        async def execute_query(keys: Iterable[TagField]) -> tuple[list[dict[str, Any]], str]:
            """Generate and execute the query against the API for the given item's cleaned ``keys``"""
            attributes = [item.clean_tags.get(key) for key in keys]
            q = " ".join(str(attr) for attr in attributes if attr)
            return await self.api.query(q, kind=kind, limit=settings.result_count), q

        results, query = await execute_query(settings.search_fields_1)
        if not results and settings.search_fields_2:
            results, query = await execute_query(settings.search_fields_2)
        if not results and settings.search_fields_3:
            results, query = await execute_query(settings.search_fields_3)

        if results:
            self.matcher.log([item.name, f"Query: {query}", f"{len(results)} results"])
            return results
        self.matcher.log([item.name, f"Query: {query}", "Match failed: No results."], pad="<")

    def _log_results(self, results: Mapping[str, ItemSearchResult]) -> None:
        """Logs the final results of the ItemSearcher"""
        if not results:
            return

        max_width = get_max_width(results)

        total_matched = 0
        total_unmatched = 0
        total_skipped = 0
        total_all = 0

        for name, result in results.items():
            matched = len(result.matched)
            unmatched = len(result.unmatched)
            skipped = len(result.skipped)
            total = matched + unmatched + skipped

            total_matched += matched
            total_unmatched += unmatched
            total_skipped += skipped
            total_all += total

            colour1 = "\33[92m" if matched > 0 else "\33[94m"
            colour2 = "\33[92m" if unmatched == 0 else "\33[91m"
            colour3 = "\33[92m" if skipped == 0 else "\33[93m"

            self.logger.report(
                f"\33[1m{align_string(name, max_width=max_width)} \33[0m|"
                f"{colour1}{matched:>6} matched \33[0m| "
                f"{colour2}{unmatched:>6} unmatched \33[0m| "
                f"{colour3}{skipped:>6} skipped \33[0m| "
                f"\33[97m{total:>6} total \33[0m"
            )

        self.logger.report(
            f"\33[1;96m{'TOTALS':<{max_width}} \33[0m|"
            f"\33[92m{total_matched:>6} matched \33[0m| "
            f"\33[91m{total_unmatched:>6} unmatched \33[0m| "
            f"\33[93m{total_skipped:>6} skipped \33[0m| "
            f"\33[97m{total_all:>6} total \33[0m"
        )
        self.logger.print_line(REPORT)

    @staticmethod
    def _determine_remote_object_type(obj: MusifyObject) -> RemoteObjectType:
        if hasattr(obj, "kind"):
            return obj.kind
        raise MusifyAttributeError(f"Given object does not specify a RemoteObjectType: {obj.__class__.__name__}")

    def __call__[T: MusifyItemSettable](
            self, collections: Collection[MusifyCollection[T]]
    ) -> Awaitable[dict[str, ItemSearchResult[T]]]:
        return self.search(collections)

    async def search[T: MusifyItemSettable](
            self, collections: Collection[MusifyCollection[T]]
    ) -> dict[str, ItemSearchResult[T]]:
        """
        Searches for remote matches for the given list of item collections.

        :return: Map of the collection's name to its :py:class:`ItemSearchResult` object.
        """
        self.logger.debug("Searching: START")
        if not [item for c in collections for item in c.items if item.has_uri is None]:
            self.logger.debug("\33[93mNo items to search. \33[0m")
            return {}

        kinds = {coll.__class__.__name__ for coll in collections}
        kind = kinds.pop() if len(kinds) == 1 else "collection"
        self.logger.info(
            f"\33[1;95m ->\33[1;97m "
            f"Searching for matches on {self.api.source} for {len(collections)} {kind}s\33[0m"
        )

        async def _get_result(coll: MusifyCollection[T]) -> tuple[str, ItemSearchResult]:
            return coll.name, await self._search_collection(coll)

        # WARNING: making this run asynchronously will break tqdm; bar will get stuck after 1-2 ticks
        bar = self.logger.get_synchronous_iterator(collections, desc="Searching",  unit=f"{kind}s")
        search_results = dict([await _get_result(coll) for coll in bar])

        self.logger.print_line()
        self._log_results(search_results)
        self.logger.debug("Searching: DONE\n")
        return search_results

    async def _search_collection[T: MusifyItemSettable](self, collection: MusifyCollection) -> ItemSearchResult[T]:
        kind = collection.__class__.__name__

        skipped = tuple(item for item in collection if item.has_uri is not None)
        if len(skipped) == len(collection):
            self.matcher.log([collection.name, "Skipping search, no items to search"], pad='<')

        if getattr(collection, "compilation", True) is False:
            self.matcher.log([collection.name, "Searching for collection as a unit"], pad='>')
            await self._search_collection_unit(collection=collection)

            missing = [item for item in collection.items if item.has_uri is None]
            if missing:
                self.matcher.log(
                    [collection.name, f"Searching for {len(missing)} unmatched items in this {kind}"]
                )
                await self._search_items(collection=collection)
        else:
            self.matcher.log([collection.name, "Searching for distinct items in collection"], pad='>')
            await self._search_items(collection=collection)

        return ItemSearchResult(
            matched=tuple(item for item in collection if item.has_uri and item not in skipped),
            unmatched=tuple(item for item in collection if item.has_uri is None and item not in skipped),
            skipped=skipped
        )

    async def _get_item_match[T: MusifyItemSettable](
            self, item: T, match_on: UnitIterable[TagField] | None = None, results: Iterable[T] = None
    ) -> tuple[T, T | None]:
        kind = self._determine_remote_object_type(item)
        search_config = self.search_settings[kind]

        if results is None:
            responses = await self._get_results(item, kind=kind, settings=search_config)
            # noinspection PyTypeChecker
            results: Iterable[T] = map(self.factory[kind], responses or ())

        result = self.matcher(
            item,
            results=results,
            match_on=match_on if match_on is not None else search_config.match_fields,
            min_score=search_config.min_score,
            max_score=search_config.max_score,
            allow_karaoke=search_config.allow_karaoke,
        ) if results else None

        return item, result

    async def _search_items[T: MusifyItemSettable](self, collection: Iterable[T], **kwargs) -> None:
        """
        Search for matches on individual items in an item collection that have ``None`` on ``has_uri`` attribute.
        kwargs are not required and are passed on to self._get_item_match.
        """
        async def _match(item: T) -> None:
            if item.has_uri is not None:
                return

            item, match = await self._get_item_match(item, **kwargs)
            if match and match.has_uri:
                item.uri = match.uri

        await self.logger.get_asynchronous_iterator(map(_match, collection), disable=True)

    async def _search_collection_unit[T: MusifyItemSettable](self, collection: MusifyCollection[T]) -> None:
        """
        Search for matches on an entire collection as a whole
        i.e. search for just the collection and not its distinct items.
        """
        if all(item.has_uri for item in collection):
            return

        kind = self._determine_remote_object_type(collection)
        search_config = self.search_settings[kind]

        responses = await self._get_results(collection, kind=kind, settings=search_config)
        key = self.api.collection_item_map[kind]
        await self.logger.get_asynchronous_iterator(
            (self.api.extend_items(response, kind=kind, key=key, leave_bar=False) for response in responses),
            disable=True
        )

        # noinspection PyProtectedMember,PyTypeChecker
        # order to prioritise results that are closer to the item count of the input collection
        results: list[T] = sorted(map(self.factory[kind], responses), key=lambda x: abs(x._total - len(collection)))

        result = self.matcher(
            collection,
            results=results,
            match_on=search_config.match_fields,
            min_score=search_config.min_score,
            max_score=search_config.max_score,
            allow_karaoke=search_config.allow_karaoke,
        )

        if not result:
            return

        # check all items in the collection have been matched
        # get matches on those that are still missing matches
        await self._search_items(collection, match_on=[Tag.TITLE], results=result.items)

    def as_dict(self) -> dict[str, Any]:
        return {
            "matcher": self.matcher,
            "remote_source": self.factory.api.source,
        }



"""
Processors that filter down objects and data types based on some given configuration.
"""
from __future__ import annotations

import logging
from collections.abc import Collection
from dataclasses import field, dataclass
from typing import Any

from musify.base import Result
from musify.field import TagField
from musify.logger import MusifyLogger
from musify.processors.base import Filter, FilterComposite
from musify.processors.filter import FilterComparers, FilterDefinedList


@dataclass(frozen=True)
class MatchResult[T: Any](Result):
    """Results from :py:class:`FilterMatcher` separated by individual filter results."""
    #: Objects that matched include settings.
    included: Collection[T] = field(default=tuple())
    #: Objects that matched exclude settings.
    excluded: Collection[T] = field(default=tuple())
    #: Objects that matched :py:class:`Comparer` settings
    compared: Collection[T] = field(default=tuple())
    #: Objects that matched on any ``group_by`` settings
    grouped: Collection[T] = field(default=tuple())

    @property
    def combined(self) -> list[T]:
        """Combine the individual results to one combined list"""
        return [track for track in [*self.compared, *self.included, *self.grouped] if track not in self.excluded]


class FilterMatcher[T: Any, U: Filter, V: Filter, X: FilterComparers](FilterComposite[T]):
    """
    Get matches for items based on given filters.

    :param include: A Filter for simple include comparisons to use when matching.
    :param exclude: A Filter for simple exclude comparisons to use when matching.
    :param comparers: A Filter for fine-grained comparisons to use when matching.
        When not given or the given Filter is not ready,
        returns all given values on match unless include or exclude are defined and ready.
    :param group_by: Once all other filters are applied, also include all other items that match this tag type
        from the matched items for any remaining unmatched items.
    """

    __slots__ = ("logger", "include", "exclude", "comparers", "group_by")

    def __init__(
            self,
            include: U = FilterDefinedList(),
            exclude: V = FilterDefinedList(),
            comparers: X = FilterComparers(),
            group_by: TagField | None = None,
            *_,
            **__
    ):
        super().__init__(comparers, include, exclude)
        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

        #: The filter that, when processed, returns items to include
        self.include = include
        #: The filter that, when processed, returns items to exclude
        self.exclude = exclude
        #: The comparers to use when processing for this filter
        self.comparers = comparers
        #: Once all other filters are applied, also include all other items that match this tag type
        #: from the matched items for the remaining items given
        self.group_by = group_by

    def __call__(self, *args, **kwargs) -> list[T]:
        return self.process(*args, **kwargs)

    def process(self, values: Collection[T], reference: T | None = None, *_, **__) -> list[T]:
        """
        Return a new, filtered list of items from input ``values`` that match the stored filters.

        :param values: List of items to filter.
        :param reference: Optional reference track to use when filtering on
            comparers and the comparer has no expected value.
        :return: List of items that match the conditions.
        """
        return self.process_to_result(values=values, reference=reference).combined

    def process_to_result(self, values: Collection[T], reference: T | None = None, *_, **__) -> MatchResult:
        """Same as :py:meth:`process` but returns the results of each filter to a :py:class`MatchResult` object"""
        if len(values) == 0:  # skip match
            return MatchResult()

        included = self.include(values)
        excluded = self.exclude(values) if self.exclude.ready else ()
        tracks_reduced = {track for track in values if track not in included}
        compared = self.comparers(tracks_reduced, reference=reference) if self.comparers.ready else ()

        result = MatchResult(included=included, excluded=excluded, compared=compared)
        grouped = self._get_group_by_results(values, matched=result.combined)

        if not grouped:
            return result
        return MatchResult(included=included, excluded=excluded, compared=compared, grouped=grouped)

    def _get_group_by_results(self, values: Collection[T], matched: Collection[T]) -> tuple[T, ...]:
        if not self.group_by or len(values) == len(matched):
            return ()
        tag_names = self.group_by.to_tag()
        tag_values = {item[tag_name] for item in matched for tag_name in tag_names if hasattr(item, tag_name)}

        return tuple(
            item for item in values
            if item not in matched
            and any(item[tag_name] in tag_values for tag_name in tag_names if hasattr(item, tag_name))
        )

    def as_dict(self):
        return {
            "include": self.include,
            "exclude": self.exclude,
            "comparers": self.comparers,
            "group_by": self.group_by.name.lower() if self.group_by else None
        }



"""
Operations relating to reading and writing tags/metadata/properties to various types of audio files.

Specific audio file types should implement :py:class:`LocalTrack`.
"""
from ._tags.writer import SyncResultTrack
from .flac import FLAC
from .m4a import M4A
from .mp3 import MP3
from .track import LocalTrack
from .utils import TRACK_CLASSES, TRACK_FILETYPES, load_track
from .wma import WMA



"""
The core abstract implementations of :py:class:`MusifyItem` and :py:class:`MusifyCollection` classes.
"""
from __future__ import annotations

import datetime
from abc import ABCMeta, abstractmethod
from collections.abc import Collection, Mapping, Iterable
from copy import deepcopy
from pathlib import Path
from typing import Self

from yarl import URL

from musify.base import MusifyItem, HasLength
from musify.exception import MusifyTypeError
from musify.libraries.core.collection import MusifyCollection
from musify.libraries.remote.core.types import RemoteObjectType


class Track(MusifyItem, HasLength, metaclass=ABCMeta):
    """Represents a track including its metadata/tags/properties."""

    __slots__ = ()
    __attributes_ignore__ = ("name",)

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def kind(cls) -> RemoteObjectType:
        """The type of remote object associated with this class"""
        return RemoteObjectType.TRACK

    @property
    def name(self) -> str:
        """This track's title"""
        return self.title

    @property
    @abstractmethod
    def title(self) -> str | None:
        """This track's title"""
        raise NotImplementedError

    @property
    @abstractmethod
    def artist(self) -> str | None:
        """Joined string representation of all artists featured on this track"""
        raise NotImplementedError

    @property
    @abstractmethod
    def artists(self) -> list[str | Artist]:
        """List of all artists featured on this track."""
        raise NotImplementedError

    @property
    @abstractmethod
    def album(self) -> str | None:
        """The album this track is featured on"""
        raise NotImplementedError

    @property
    @abstractmethod
    def album_artist(self) -> str | None:
        """The artist of the album this track is featured on"""
        raise NotImplementedError

    @property
    @abstractmethod
    def track_number(self) -> int | None:
        """The position this track has on the album it is featured on"""
        raise NotImplementedError

    @property
    @abstractmethod
    def track_total(self) -> int | None:
        """The track number of tracks on the album this track is featured on"""
        raise NotImplementedError

    @property
    @abstractmethod
    def genres(self) -> list[str] | None:
        """List of genres associated with this track"""
        raise NotImplementedError

    @property
    def date(self) -> datetime.date | None:
        """A :py:class:`date` object representing the release date of this track"""
        if self.year and self.month and self.day:
            return datetime.date(self.year, self.month, self.day)

    @property
    @abstractmethod
    def year(self) -> int | None:
        """The year this track was released"""
        raise NotImplementedError

    @property
    @abstractmethod
    def month(self) -> int | None:
        """The month this track was released"""
        raise NotImplementedError

    @property
    @abstractmethod
    def day(self) -> int | None:
        """The day this track was released"""
        raise NotImplementedError

    @property
    @abstractmethod
    def bpm(self) -> float | None:
        """The tempo of this track"""
        raise NotImplementedError

    @property
    @abstractmethod
    def key(self) -> str | None:
        """The key of this track in alphabetical musical notation format"""
        raise NotImplementedError

    @property
    @abstractmethod
    def disc_number(self) -> int | None:
        """The number of the disc from the album this track is featured on"""
        raise NotImplementedError

    @property
    @abstractmethod
    def disc_total(self) -> int | None:
        """The total number the discs from the album this track is featured on"""
        raise NotImplementedError

    @property
    @abstractmethod
    def compilation(self) -> bool | None:
        """Is the album this track is featured on a compilation"""
        raise NotImplementedError

    @property
    @abstractmethod
    def comments(self) -> list[str] | None:
        """Comments associated with this track set by the user"""
        raise NotImplementedError

    @property
    @abstractmethod
    def image_links(self) -> dict[str, str | Path | URL]:
        """
        The images associated with the album this track is featured on in the form ``{<image name/type>: <image link>}``
        """
        raise NotImplementedError

    @property
    def has_image(self) -> bool:
        """Does the album this track is associated with have an image"""
        return len(self.image_links) > 0

    @property
    @abstractmethod
    def length(self) -> float | None:
        """Total duration of this track in seconds"""
        raise NotImplementedError

    @property
    @abstractmethod
    def rating(self) -> float | None:
        """The rating for this track"""
        raise NotImplementedError


class Playlist[T: Track](MusifyCollection[T], metaclass=ABCMeta):
    """A playlist of items and their derived properties/objects."""

    __slots__ = ()
    __attributes_classes__ = MusifyCollection
    __attributes_ignore__ = ("items",)

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def kind(cls) -> RemoteObjectType:
        """The type of remote object associated with this class"""
        return RemoteObjectType.PLAYLIST

    @property
    @abstractmethod
    def name(self):
        """The name of this playlist"""
        raise NotImplementedError

    @property
    @abstractmethod
    def description(self) -> str | None:
        """Description of this playlist"""
        raise NotImplementedError

    @property
    def items(self):
        """The tracks in this collection"""
        return self.tracks

    @property
    @abstractmethod
    def tracks(self) -> list[T]:
        """The tracks in this playlist"""
        raise NotImplementedError

    @property
    def track_total(self) -> int:
        """The total number of tracks in this playlist"""
        return len(self)

    @property
    @abstractmethod
    def image_links(self) -> dict[str, str]:
        """The images associated with this playlist in the form ``{<image name/type>: <image link>}``"""
        raise NotImplementedError

    @property
    def has_image(self) -> bool:
        """Does this playlist have an image"""
        return len(self.image_links) > 0

    @property
    def length(self):
        """Total duration of all tracks in this playlist in seconds"""
        lengths = {track.length for track in self.tracks}
        return sum(lengths) if lengths else None

    @property
    @abstractmethod
    def date_created(self) -> datetime.datetime | None:
        """:py:class:`datetime.datetime` object representing when the playlist was created"""
        raise NotImplementedError

    @property
    @abstractmethod
    def date_modified(self) -> datetime.datetime | None:
        """:py:class:`datetime.datetime` object representing when the playlist was last modified"""
        raise NotImplementedError

    def merge(self, other: Iterable[T], reference: Self | None = None) -> None:
        """
        Merge tracks in this playlist with another collection, synchronising tracks between the two.
        Only modifies this playlist.

        Sort order is not preserved when merging.
        Any items that need to be added to this playlist will be added at the end of the playlist.
        Duplicates that are present in the ``other`` collection are filtered out by default.

        :param other: The collection of items to merge onto this playlist.
        :param reference: Optionally, provide a reference playlist to compare both the current playlist
            and the ``other`` items to. The function will determine tracks to remove from
            this playlist based on the reference. Useful for using this function as a synchronizer
            where the reference refers to the playlist at the previous sync.
        """
        if not self._validate_item_type(other):
            raise MusifyTypeError([type(i).__name__ for i in other])

        if reference is None:
            self.extend(self.outer_difference(other), allow_duplicates=False)
            return

        for item in reference:
            if item not in other and item in self:
                self.remove(item)

        self.extend(reference.outer_difference(other), allow_duplicates=False)

    def __or__(self, other: Playlist[T]) -> Self:
        if not isinstance(other, self.__class__):
            raise MusifyTypeError(
                f"Incorrect item given. Cannot merge with {other.__class__.__name__} "
                f"as it is not a {self.__class__.__name__}"
            )

        self_copy = deepcopy(self)
        self_copy.merge(other)
        return self_copy

    def __ior__(self, other: Playlist[T]) -> Self:
        if not isinstance(other, self.__class__):
            raise MusifyTypeError(
                f"Incorrect item given. Cannot merge with {other.__class__.__name__} "
                f"as it is not a {self.__class__.__name__}"
            )

        self.merge(other)
        return self


type LibraryMergeType[T] = Library[T] | Collection[Playlist[T]] | Mapping[str, Playlist[T]]


class Library[T: Track](MusifyCollection[T], metaclass=ABCMeta):
    """A library of items and playlists and other object types."""

    __slots__ = ()
    __attributes_classes__ = MusifyCollection
    __attributes_ignore__ = ("items",)

    @property
    @abstractmethod
    def name(self):
        """The library name"""
        raise NotImplementedError

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def source(cls) -> str:
        """The type of library loaded"""
        return cls.__name__.replace("Library", "")

    @property
    def items(self):
        """The tracks in this collection"""
        return self.tracks

    @property
    def length(self):
        """Total duration of all tracks in this library in seconds"""
        lengths = {track.length for track in self.tracks}
        return sum(lengths) if lengths else None

    @property
    @abstractmethod
    def tracks(self) -> list[T]:
        """The tracks in this library"""
        raise NotImplementedError

    @property
    def track_total(self) -> int:
        """The total number of tracks in this library"""
        return len(self)

    @property
    def tracks_in_playlists(self) -> set[T]:
        """All unique tracks from all playlists in this library"""
        return {track for pl in self.playlists.values() for track in pl}

    @property
    @abstractmethod
    def playlists(self) -> dict[str, Playlist[T]]:
        """The playlists in this library"""
        raise NotImplementedError

    @abstractmethod
    async def load(self):
        """Implementations of this function should load all data for this library and log results."""
        raise NotImplementedError

    @abstractmethod
    async def load_tracks(self) -> None:
        """
        Implementations of this function should load all tracks for this library
        and store them within the library object to be retrieved with property ``tracks``.
        """
        raise NotImplementedError

    @abstractmethod
    def log_tracks(self) -> None:
        """Log stats on currently loaded tracks"""
        raise NotImplementedError

    @abstractmethod
    async def load_playlists(self) -> None:
        """
        Implementations of this function should load all playlists for this library
        and store them within the library object to be retrieved with property ``playlists``.
        """
        raise NotImplementedError

    @abstractmethod
    def log_playlists(self) -> None:
        """Log stats on currently loaded playlists"""
        raise NotImplementedError

    def merge_playlists(self, playlists: LibraryMergeType[T], reference: LibraryMergeType[T] | None = None) -> None:
        """
        Merge playlists from given list/map/library to this library.

        See :py:meth:`.Playlist.merge` for more info.

        :param playlists: The playlists to merge onto this library's playlists.
            If a given playlist is not found in this library, simply add the playlist to this library.
        :param reference: Optionally, provide a reference playlist to compare both the current playlist
            and the ``other`` items to. The function will determine tracks to remove from
            this playlist based on the reference. Useful for using this function as a synchronizer
            where the reference refers to the playlist at the previous sync.
        """
        def get_playlists_map(value: LibraryMergeType[T]) -> Mapping[str, Playlist[T]]:
            """Reformat the input playlist values to map"""
            if isinstance(value, Mapping):
                return value
            elif isinstance(value, Library):
                return value.playlists
            elif isinstance(value, Collection):
                return {pl.name: pl for pl in value}
            raise MusifyTypeError(f"Unrecognised input type: {value.__class__.__name__}")

        playlists = get_playlists_map(playlists)
        reference = get_playlists_map(reference) if reference is not None else {}

        for name, playlist in playlists.items():
            if name not in self.playlists:
                self.playlists[name] = deepcopy(playlist)
                continue

            self.playlists[name].merge(playlist, reference=reference.get(name))


class Folder[T: Track](MusifyCollection[T], metaclass=ABCMeta):
    """
    A folder of items and their derived properties/objects
    """

    __slots__ = ()
    __attributes_classes__ = MusifyCollection
    __attributes_ignore__ = ("name", "items")

    @property
    @abstractmethod
    def name(self):
        """The folder name"""
        raise NotImplementedError

    @property
    def folder(self) -> str:
        """The folder name"""
        return self.name

    @property
    def items(self):
        """The tracks in this collection"""
        return self.tracks

    @property
    @abstractmethod
    def tracks(self):
        """The tracks in this folder"""
        raise NotImplementedError

    @property
    @abstractmethod
    def artists(self) -> list[str]:
        """List of artists ordered by frequency of appearance on the tracks in this folder"""
        raise NotImplementedError

    @property
    @abstractmethod
    def albums(self) -> list[str]:
        """List of albums ordered by frequency of appearance on the tracks in this folder"""
        raise NotImplementedError

    @property
    def track_total(self) -> int:
        """The total number of tracks in this folder"""
        return len(self)

    @property
    @abstractmethod
    def genres(self) -> list[str]:
        """List of genres ordered by frequency of appearance on the tracks in this folder"""
        raise NotImplementedError

    @property
    @abstractmethod
    def compilation(self) -> bool:
        """Is this folder a compilation"""
        raise NotImplementedError

    @property
    def length(self):
        """Total duration of all tracks in this folder in seconds"""
        lengths = {track.length for track in self.tracks}
        return sum(lengths) if lengths else None


class Album[T: Track](MusifyCollection[T], metaclass=ABCMeta):
    """An album of items and their derived properties/objects."""

    __slots__ = ()
    __attributes_classes__ = MusifyCollection
    __attributes_ignore__ = ("name", "items")

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def kind(cls) -> RemoteObjectType:
        """The type of remote object associated with this class"""
        return RemoteObjectType.ALBUM

    @property
    @abstractmethod
    def name(self) -> str:
        """The album name"""
        raise NotImplementedError

    @property
    def items(self):
        """The tracks in this collection"""
        return self.tracks

    @property
    def album(self) -> str:
        """The album name"""
        return self.name

    @property
    @abstractmethod
    def tracks(self) -> list[T]:
        """The tracks on this album"""
        raise NotImplementedError

    @property
    def artist(self) -> str:
        """Joined string representation of all artists on this album ordered by frequency of appearance"""
        return self.tag_sep.join(self.artists)

    @property
    @abstractmethod
    def artists(self) -> list[str | Artist]:
        """List of artists ordered by frequency of appearance on the tracks on this album"""
        raise NotImplementedError

    @property
    @abstractmethod
    def album_artist(self) -> str | None:
        """The album artist for this album"""
        raise NotImplementedError

    @property
    def track_total(self) -> int:
        """The total number of tracks on this album"""
        return len(self)

    @property
    @abstractmethod
    def genres(self) -> list[str]:
        """List of genres ordered by frequency of appearance on the tracks on this album"""
        raise NotImplementedError

    @property
    def date(self) -> datetime.date | None:
        """A :py:class:`date` object representing the release date of this album"""
        if self.year and self.month and self.day:
            return datetime.date(self.year, self.month, self.day)

    @property
    @abstractmethod
    def year(self) -> int | None:
        """The year this album was released"""
        raise NotImplementedError

    @property
    @abstractmethod
    def month(self) -> int | None:
        """The month this album was released"""
        raise NotImplementedError

    @property
    @abstractmethod
    def day(self) -> int | None:
        """The day this album was released"""
        raise NotImplementedError

    @property
    def disc_total(self) -> int | None:
        """The highest value of disc number on this album"""
        disc_numbers = {track.disc_number for track in self.tracks if track.disc_number}
        return max(disc_numbers) if disc_numbers else None

    @property
    @abstractmethod
    def compilation(self) -> bool:
        """Is this album a compilation"""
        raise NotImplementedError

    @property
    @abstractmethod
    def image_links(self) -> dict[str, str]:
        """The images associated with this album in the form ``{<image name/type>: <image link>}``"""
        raise NotImplementedError

    @property
    def has_image(self) -> bool:
        """Does this album have an image"""
        return len(self.image_links) > 0

    @property
    def length(self):
        """Total duration of all tracks on this album in seconds"""
        lengths = {track.length for track in self.tracks}
        return sum(lengths) if lengths else None

    @property
    @abstractmethod
    def rating(self) -> float | None:
        """Rating of this album"""
        raise NotImplementedError


class Artist[T: (Track, Album)](MusifyCollection[T], metaclass=ABCMeta):
    """An artist of items and their derived properties/objects."""

    __slots__ = ()
    __attributes_classes__ = MusifyCollection
    __attributes_ignore__ = ("name", "items")

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def kind(cls) -> RemoteObjectType:
        """The type of remote object associated with this class"""
        return RemoteObjectType.ARTIST

    @property
    @abstractmethod
    def name(self):
        """The artist name"""
        raise NotImplementedError

    @property
    def items(self):
        """The tracks in this collection"""
        return self.tracks

    @property
    def artist(self) -> str:
        """The artist name"""
        return self.name

    @property
    @abstractmethod
    def tracks(self) -> list[T]:
        """The tracks by this artist"""
        raise NotImplementedError

    @property
    @abstractmethod
    def artists(self) -> list[str]:
        """List of other artists ordered by frequency of appearance on the albums by this artist"""
        raise NotImplementedError

    @property
    @abstractmethod
    def albums(self) -> list[str | Album]:
        """List of albums ordered by frequency of appearance on the tracks by this artist"""
        raise NotImplementedError

    @property
    def track_total(self) -> int:
        """The total number of tracks by this artist"""
        return len(self)

    @property
    @abstractmethod
    def genres(self) -> list[str]:
        """List of genres for this artist"""
        raise NotImplementedError

    @property
    def length(self):
        """Total duration of all tracks by this artist in seconds"""
        lengths = {track.length for track in self.tracks}
        return sum(lengths) if lengths else None

    @property
    @abstractmethod
    def rating(self) -> int | None:
        """The popularity of this artist"""
        raise NotImplementedError


class Genre[T: Track](MusifyCollection[T], metaclass=ABCMeta):
    """A genre of items and their derived properties/objects."""

    __slots__ = ()
    __attributes_classes__ = MusifyCollection
    __attributes_ignore__ = ("name", "items")

    @property
    @abstractmethod
    def name(self):
        """The genre"""
        raise NotImplementedError

    @property
    def items(self):
        """The tracks in this collection"""
        return self.tracks

    @property
    def genre(self) -> str:
        """The genre"""
        return self.name

    @property
    @abstractmethod
    def tracks(self) -> list[T]:
        """The tracks for this genre"""
        raise NotImplementedError

    @property
    @abstractmethod
    def artists(self) -> list[str]:
        """List of artists ordered by frequency of appearance on the tracks for this genre"""
        raise NotImplementedError

    @property
    @abstractmethod
    def albums(self) -> list[str]:
        """List of albums ordered by frequency of appearance on the tracks for this genre"""
        raise NotImplementedError

    @property
    @abstractmethod
    def related_genres(self) -> list[str]:
        """List of related genres ordered by frequency of appearance on the tracks for this genre"""
        raise NotImplementedError

    @property
    def length(self):
        """Total duration of all tracks with this genre in seconds"""
        lengths = {track.length for track in self.tracks}
        return sum(lengths) if lengths else None



"""
The fundamental core printer classes for the entire package.
"""
import re
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping
from concurrent.futures import Future, ThreadPoolExecutor
from datetime import datetime, date
from pathlib import Path
from typing import Any

from aiorequestful.types import ImmutableJSON, JSON, JSON_VALUE
from yarl import URL

from musify.types import UnitIterable
from musify.utils import to_collection


class PrettyPrinter(metaclass=ABCMeta):
    """Generic base class for pretty printing. Classes can inherit this class to gain pretty print functionality."""

    __slots__ = ()

    _upper_key_words = {"id", "uri", "url", "bpm"}
    _max_val_width = 120

    @staticmethod
    def _pascal_to_snake(value: str) -> str:
        """Convert PascalCase to snake_case."""
        value = re.sub(r"([A-Z])", lambda m: f"_{m.group(1).lower()}", value.strip("_ "))
        value = re.sub(r"[_ ]+", "_", value).strip("_ ")
        return value.lower()

    @staticmethod
    def _snake_to_pascal(value: str) -> str:
        """Convert snake_case to PascalCase."""
        value = re.sub(r"_(.)", lambda m: m.group(1).upper(), value.strip().lower())
        return re.sub(r"^(.)", lambda m: m.group(1).upper(), value.strip())

    @abstractmethod
    def as_dict(self) -> dict[str, Any]:
        """
        Return a dictionary representation of the key attributes of this object.

        The results of this function are used to produce the following:
            * A JSON representation of the object when calling :py:func:`json`
            * The string representation of the object when calling str() on the object
            * The representation of the object when calling repr() on the object
        """
        raise NotImplementedError

    def _json_attributes(self):
        return self.as_dict()

    def json(self) -> JSON:
        """Return a dictionary representation of the key attributes of this object that is safe to output to JSON"""
        return self._to_json(self._json_attributes())

    @classmethod
    def _to_json(cls, attributes: ImmutableJSON, pool: bool = False) -> JSON:
        def _get_json_key_value(attribute: tuple[Any, Any]) -> tuple[str, JSON_VALUE | list[Future[JSON_VALUE]]]:
            key, value = attribute
            return str(key), cls._get_json_value(value=value)

        if not pool:
            tasks = map(_get_json_key_value, attributes.items())
        else:
            with ThreadPoolExecutor(thread_name_prefix="to-json") as executor:
                tasks = executor.map(_get_json_key_value, attributes.items())

        return dict(tasks)

    @classmethod
    def _get_json_value(cls, value: Any, pool: bool = False) -> JSON_VALUE | list[Future[JSON_VALUE]]:
        if isinstance(value, set):
            value = to_collection(value)

        if isinstance(value, (list, tuple)):
            if not pool:
                tasks = map(cls._get_json_value, value)
            else:
                with ThreadPoolExecutor(thread_name_prefix="to-json-value") as executor:
                    tasks = executor.map(cls._get_json_value, value)
            return list(tasks)
        elif isinstance(value, Mapping):
            return cls._to_json(value, pool=pool)
        elif isinstance(value, PrettyPrinter):
            return value._to_json(value._json_attributes(), pool=pool)
        elif isinstance(value, (datetime, date, Path, URL)):
            return str(value)

        return value

    def __str__(self, indent: int = 2, increment: int = 2):
        obj_dict = self.as_dict()
        if not obj_dict:
            return f"{self.__class__.__name__}()"

        result = f"{self.__class__.__name__}(\n{{}}\n" + " " * (indent - increment) + ")"
        attributes_repr = self._to_str(obj_dict, indent=indent, increment=increment)
        return result.format("\n".join([" " * indent + attribute for attribute in attributes_repr]))

    @classmethod
    def _to_str(cls, attributes: Mapping[str, Any], indent: int = 2, increment: int = 2) -> list[str]:
        if len(attributes) == 0:
            return []
        max_key_width = max(map(len, map(str, attributes))) + 1  # +1 for space after key
        max_key_width += max_key_width % increment
        max_val_width = cls._max_val_width - max_key_width

        indent_prev = indent
        indent += increment

        attributes_repr: list[str] = []
        for attr_key, attr_val in attributes.items():
            attr_key = str(attr_key).title().replace('_', ' ')
            for word in cls._upper_key_words:
                pattern = re.compile(rf"(^{word}$|^{word} | {word}$| {word})", flags=re.I)
                attr_key = re.sub(pattern, lambda m: m.group(1).upper(), attr_key)

            if isinstance(attr_val, PrettyPrinter):
                attr_val_str = attr_val.__str__(indent=indent, increment=increment)
            elif isinstance(attr_val, (datetime, date)):
                attr_val_str = str(attr_val)
            elif isinstance(attr_val, (list, tuple, set)) and len(attr_val) > 0:
                attr_val_str = cls._get_attribute_value_from_collection(
                    attr_val=attr_val,
                    max_val_width=max_val_width,
                    indent=indent,
                    indent_prev=indent_prev,
                    increment=increment
                )
            elif isinstance(attr_val, Mapping) and len(attr_val) > 0:
                attr_val_str = cls._get_attribute_value_from_mapping(
                    attr_val=attr_val,
                    max_val_width=max_val_width,
                    indent=indent,
                    indent_prev=indent_prev,
                    increment=increment
                )
            else:
                attr_val_str = repr(attr_val)

            attr_val_repr = f"{attr_key} = {attr_val_str}"  # for single line format
            if len(attr_val_repr) > max_val_width or len(attributes) > 1:
                # reformat to indented list format
                attr_val_repr = f"{attr_key: <{max_key_width}}= {attr_val_str}"
            attributes_repr.append(attr_val_repr)

        return attributes_repr

    @staticmethod
    def _get_attribute_value_from_collection(
            attr_val: list | tuple | set, max_val_width: int, indent: int, indent_prev: int, increment: int
    ) -> str:
        pp_repr = "[{}]"
        if isinstance(attr_val, set):
            pp_repr = "{{}}"
        elif isinstance(attr_val, tuple):
            pp_repr = "({})"

        attr_val_str = str(attr_val)

        if len(attr_val_str) > max_val_width:
            pp_repr = pp_repr.format("\n" + " " * indent + "{}\n" + " " * indent_prev)
            attr_val_pp = []
            for val in attr_val:
                if isinstance(val, PrettyPrinter):
                    attr_val_pp.append(val.__str__(indent=indent + increment, increment=increment))
                else:
                    attr_val_pp.append(str(val))
            attr_val_str = pp_repr.format((",\n" + " " * indent).join(attr_val_pp))

        return attr_val_str

    @classmethod
    def _get_attribute_value_from_mapping(
            cls, attr_val: Mapping, max_val_width: int, indent: int, indent_prev: int, increment: int
    ) -> str:
        attr_val_pp = cls._to_str(attr_val, indent=indent, increment=increment)
        attr_val_str = "{" + ", ".join(attr_val_pp) + "}"

        if len(attr_val_str) > max_val_width:
            pp_repr = "\n" + " " * indent + "{}\n" + " " * indent_prev
            attr_val_str = "{" + pp_repr.format((",\n" + " " * indent).join(attr_val_pp)) + "}"

        return attr_val_str

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.as_dict())})"


class AttributePrinter(PrettyPrinter):
    """
    Extends the functionality of a :py:class:`PrettyPrinter`.

    Adds functionality to automatically determine the key attributes that represent child objects
    and uses these for printer representations.
    """

    __slots__ = ()
    __attributes_classes__: UnitIterable[type] = ()
    __attributes_ignore__: UnitIterable[str] = ()

    def _get_attributes(self) -> dict[str, Any]:
        """Returns the key attributes of the current instance for pretty printing"""
        def get_settings(kls: type) -> None:
            """Build up classes and exclude keys for getting attributes"""
            if kls != self.__class__ and kls not in classes:
                classes.append(kls)
            if issubclass(kls, AttributePrinter):
                ignore.update(to_collection(kls.__attributes_ignore__))
                for k in to_collection(kls.__attributes_classes__):
                    get_settings(k)

        classes: list[type] = []
        ignore: set[str] = set()
        get_settings(self.__class__)
        classes.insert(1, self.__class__)

        attributes = {}
        for cls in classes:
            attributes |= {
                k: getattr(self, k) for k in cls.__dict__.keys()
                if k not in ignore and isinstance(getattr(cls, k), property) and not k.startswith("_")
            }

        return attributes

    def as_dict(self) -> dict[str, Any]:
        return self._get_attributes()



"""
Utilities for operations on :py:class:`LocalTrack` types.

Generally, this will contain global variables representing all supported audio file types
and a utility function for loading the appropriate :py:class:`LocalTrack` type for a path based on its extension.
"""
from pathlib import Path

from musify.file.exception import InvalidFileType
from musify.libraries.local.track import LocalTrack
from musify.libraries.local.track.flac import FLAC
from musify.libraries.local.track.m4a import M4A
from musify.libraries.local.track.mp3 import MP3
from musify.libraries.local.track.wma import WMA
from musify.libraries.remote.core.wrangle import RemoteDataWrangler

TRACK_CLASSES = frozenset({FLAC, MP3, M4A, WMA})
TRACK_FILETYPES = frozenset(filetype for c in TRACK_CLASSES for filetype in c.valid_extensions)


async def load_track(path: str | Path, remote_wrangler: RemoteDataWrangler = None) -> LocalTrack:
    """
    Attempt to load a file from a given path, returning the appropriate :py:class:`LocalTrack` object

    :param path: The path of the file to load.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs.
        This object will be used to check for and validate a URI tag on the file.
        The tag that is used for reading and writing is set by the ``uri_tag`` class attribute on the track object.
        If no ``remote_wrangler`` is given, no URI processing will occur.
    :return: Loaded :py:class:`LocalTrack` object
    :raise InvalidFileType: If the file type is not supported.
    """
    ext = Path(path).suffix
    if ext not in TRACK_FILETYPES:
        raise InvalidFileType(ext, f"Not an accepted extension. Use only: {', '.join(TRACK_FILETYPES)}")

    cls = next(cls for cls in TRACK_CLASSES if ext in cls.valid_extensions)
    return await cls(file=path, remote_wrangler=remote_wrangler)



"""
Convert and validate remote ID and item types according to specific remote implementations.
"""
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping

from aiorequestful.types import URLInput
from yarl import URL

from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.exception import RemoteObjectTypeError
from musify.libraries.remote.core.types import APIInputValueSingle, APIInputValueMulti, RemoteIDType, RemoteObjectType


class RemoteDataWrangler(metaclass=ABCMeta):
    """Convert and validate remote ID and item types according to specific remote implementations."""

    __slots__ = ()

    @property
    @abstractmethod
    def source(self) -> str:
        """The name of the service for which data can be processed by this wrangler"""
        raise NotImplementedError

    @property
    @abstractmethod
    def unavailable_uri_dummy(self) -> str:
        """
        The value to use as a URI for an item which does not have an associated remote object.
        An item that has this URI value will be excluded from most remote logic.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def url_api(self) -> URL:
        """The base URL of the API for this remote source"""
        raise NotImplementedError

    @property
    @abstractmethod
    def url_ext(self) -> URL:
        """The base URL of external links for this remote source"""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_id_type(value: str, kind: RemoteObjectType | None = None) -> RemoteIDType:
        """
        Determine the remote ID type of the given ``value`` and return its type.

        :param value: URL/URI/ID to check.
        :param kind: When this is equal to ``USER``, ignore checks and always return ``ID`` as  type.
        :return: The :py:class:`RemoteIDType`.
        :raise RemoteIDTypeError: Raised when the function cannot determine the ID type
            of the input ``value``.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def validate_id_type(cls, value: str, kind: RemoteIDType = RemoteIDType.ALL) -> bool:
        """Check that the given ``value`` is a type of remote ID given by ``kind``"""
        raise NotImplementedError

    @classmethod
    def get_item_type(
            cls, values: APIInputValueMulti[RemoteResponse], kind: RemoteObjectType | None = None
    ) -> RemoteObjectType:
        """
        Determine the remote object type of ``values``.

        ``values`` may be:
            * A string representing a URL/URI/ID.
            * A MutableSequence of strings representing URLs/URIs/IDs of the same type.
            * A remote API JSON response for a collection including a valid item type value under a ``type`` key.
            * A MutableSequence of remote API JSON responses for a collection including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        :param values: The values representing some remote objects. See description for allowed value types.
            These items must all be of the same type of item to pass i.e. all tracks OR all artists etc.
        :param kind: The :py:class:`RemoteObjectType` to use as backup if the value is found to be an ID.
        :return: The :py:class:`RemoteObjectType`
        :raise RemoteObjectTypeError: Raised when the function cannot determine the item type
            of the input ``values``.
            Or when the list contains strings representing many differing remote object types or only IDs.
        """
        if isinstance(values, URLInput | Mapping | RemoteResponse):
            return cls._get_item_type(value=values, kind=kind)

        if len(values) == 0:
            raise RemoteObjectTypeError("No values given: collection is empty")

        kinds = {cls._get_item_type(value=value, kind=kind) for value in values if value is not None}
        kinds.discard(None)
        if len(kinds) == 0:
            raise RemoteObjectTypeError("Given items are invalid or are IDs with no kind given")
        if len(kinds) != 1:
            value = [kind.name for kind in kinds]
            raise RemoteObjectTypeError("Ensure all the given items are of the same type! Found", value=value)
        return kinds.pop()

    @staticmethod
    @abstractmethod
    def _get_item_type(
            value: APIInputValueSingle[RemoteResponse], kind: RemoteObjectType | None = None
    ) -> RemoteObjectType | None:
        """
        Determine the remote object type of the given ``value`` and return its type.

        ``value`` may be:
            * A string representing a URL/URI/ID.
            * A remote API JSON response for a collection with a valid ID value under an ``id`` key.
            * A RemoteResponse containing a remote API JSON response with the same structure as above.

        :param value: The value representing some remote collection. See description for allowed value types.
        :param kind: The :py:class:`RemoteObjectType` to use as backup if the value is found to be an ID.
        :return: The :py:class:`RemoteObjectType`. If the given value is determined to be an ID, returns None.
        :raise RemoteObjectTypeError: Raised when the function cannot determine the item type
            of the input ``values``.
        :raise EnumNotFoundError: Raised when the item type of the given ``value`` is not
            a valid remote object type.
        """
        raise NotImplementedError

    @classmethod
    def validate_item_type(cls, values: APIInputValueMulti[RemoteResponse], kind: RemoteObjectType) -> None:
        """
        Check that the given ``values`` are a type of item given by ``kind`` or a simple ID.

        ``values`` may be:
            * A string representing a URL/URI/ID.
            * A MutableSequence of strings representing URLs/URIs/IDs of the same type.
            * A remote API JSON response for a collection including a valid item type value under a ``type`` key.
            * A MutableSequence of remote API JSON responses for a collection including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        :param values: The values representing some remote objects. See description for allowed value types.
            These items must all be of the same type of item to pass i.e. all tracks OR all artists etc.
        :param kind: The remote object type to check for.
        :raise RemoteObjectTypeError: Raised when the function cannot validate the item type
            of the input ``values`` is of type ``kind`` or a simple ID.
        """
        item_type = cls.get_item_type(values, kind=kind)
        if item_type is not None and not item_type == kind:
            item_str = "unknown" if item_type is None else item_type.name.lower() + "s"
            raise RemoteObjectTypeError(f"Given items must all be {kind.name.lower()} URLs/URIs/IDs, not {item_str}")

    @classmethod
    @abstractmethod
    def convert(
            cls,
            value: str,
            kind: RemoteObjectType | None = None,
            type_in: RemoteIDType = RemoteIDType.ALL,
            type_out: RemoteIDType = RemoteIDType.ID
    ) -> str:
        """
        Converts ID to required format - API URL, EXT URL, URI, or ID.

        :param value: URL/URI/ID to convert.
        :param kind: Optionally, give the item type of the input ``value`` to skip some checks.
            This is required when the given ``value`` is an ID.
        :param type_in: Optionally, give the ID type of the input ``value`` to skip some checks.
        :param type_out: The ID type of the output ``value``.
        :return: Formatted string.
        :raise RemoteIDTypeError: Raised when the function cannot determine the item type
            of the input ``value``.
        :raise RemoteObjectTypeError: Raised when the function cannot process the input ``value``
            as a type of the given ``type_in`` i.e. the ``type_in`` does not match the actual type of the ``value``.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def extract_ids(cls, values: APIInputValueMulti[RemoteResponse], kind: RemoteObjectType | None = None) -> list[str]:
        """
        Extract a list of IDs from input ``values``.

        ``values`` may be:
            * A string representing a URL/URI/ID.
            * A MutableSequence of strings representing URLs/URIs/IDs of the same type.
            * A remote API JSON response for a collection including:
                - a valid ID value under an ``id`` key,
                - a valid item type value under a ``type`` key if ``kind`` is None.
            * A MutableSequence of remote API JSON responses for a collection including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        :param values: The values representing some remote objects. See description for allowed value types.
            These items may be of mixed item types e.g. some tracks AND some artists.
        :param kind: Optionally, give the item type of the input ``value`` to skip some checks.
            This is required when the given ``value`` is an ID.
        :return: List of IDs.
        :raise RemoteError: Raised when the function cannot determine the item type of the input ``values``.
            Or when it does not recognise the type of the input ``values`` parameter.
        """
        raise NotImplementedError



"""
Basic concrete implementation of a MusifyCollection.
"""
from __future__ import annotations

from collections.abc import Iterable, Collection
from typing import Any

from musify.base import MusifyItem
from musify.libraries.core.collection import MusifyCollection
from musify.utils import to_collection


class BasicCollection[T: MusifyItem](MusifyCollection[T]):
    """
    A basic implementation of MusifyCollection for storing ``items`` with a given ``name``.

    :param name: The name of this collection.
    :param items: The items in this collection
    """

    __slots__ = ("_name", "_items")

    @staticmethod
    def _validate_item_type(items: Any | Iterable[Any]) -> bool:
        if isinstance(items, Iterable):
            return all(isinstance(item, MusifyItem) for item in items)
        return isinstance(items, MusifyItem)

    @property
    def name(self):
        """The name of this collection"""
        return self._name

    @property
    def items(self) -> list[T]:
        return self._items

    @property
    def length(self):
        lengths = {getattr(item, "length", None) for item in self.items}
        return sum({length for length in lengths if length}) if lengths else None

    def __init__(self, name: str, items: Collection[T]):
        super().__init__()
        self._name = name
        self._items = to_collection(items, list)



"""
Implements all collection types for a local library.
"""
from __future__ import annotations

import asyncio
import logging
import sys
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping, Collection, Iterable, Container
from datetime import datetime
from pathlib import Path
from typing import Any, Self

from musify.field import Fields, TagField, TagFields
from musify.file.exception import UnexpectedPathError
from musify.libraries.core.collection import MusifyCollection
from musify.libraries.core.object import Track, Library, Folder, Album, Artist, Genre
from musify.libraries.local.base import LocalItem
from musify.libraries.local.exception import LocalCollectionError
from musify.libraries.local.track import LocalTrack, SyncResultTrack, load_track, TRACK_FILETYPES
from musify.libraries.local.track.field import LocalTrackField
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.logger import MusifyLogger
from musify.types import UnitCollection, UnitIterable
from musify.utils import get_most_common_values, to_collection, align_string, get_max_width

_max_str = "z" * 50


class LocalCollection[T: LocalTrack](MusifyCollection[T], metaclass=ABCMeta):
    """
    Generic class for storing a collection of local tracks.

    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on items.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
    """

    __slots__ = ("logger", "remote_wrangler")
    __attributes_ignore__ = ("track_total", "last_modified")

    @staticmethod
    def _validate_item_type(items: Any | Iterable[Any]) -> bool:
        if isinstance(items, Iterable):
            return all(isinstance(item, LocalItem) for item in items)
        return isinstance(items, LocalItem)

    @property
    def items(self) -> list[T]:
        """The tracks in this collection"""
        return self.tracks

    @property
    @abstractmethod
    def tracks(self) -> list[T]:
        """The tracks in this collection"""
        raise NotImplementedError

    @property
    def track_total(self) -> int:
        """The total number of tracks in this collection"""
        return len(self)

    @property
    def last_modified(self) -> datetime:
        """Timestamp of the last modified track in this collection"""
        sort = sorted(filter(lambda t: t.date_modified, self.tracks), key=lambda t: t.date_modified, reverse=True)
        # None condition will only be returned when using dummy tracks in tests therefore not included in typing
        return sort[0].date_modified if sort else None

    @property
    def last_added(self) -> datetime | None:
        """Timestamp of the track last added to the library in this collection"""
        sort = sorted(filter(lambda t: t.date_added, self.tracks), key=lambda t: t.date_added, reverse=True)
        return sort[0].date_added if sort else None

    @property
    def last_played(self) -> datetime | None:
        """Timestamp of the last played track in this collection"""
        sort = sorted(filter(lambda t: t.last_played, self.tracks), key=lambda t: t.last_played, reverse=True)
        return sort[0].last_played if sort else None

    @property
    def play_count(self) -> int:
        """Total number of plays of all tracks in this collection"""
        return sum(track.play_count for track in self.tracks if track.play_count)

    def __init__(self, remote_wrangler: RemoteDataWrangler = None):
        super().__init__()

        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)
        #: A :py:class:`RemoteDataWrangler` object for processing remote data
        self.remote_wrangler = remote_wrangler

    async def save_tracks(
            self,
            tags: UnitIterable[LocalTrackField] = LocalTrackField.ALL,
            replace: bool = False,
            dry_run: bool = True
    ) -> dict[T, SyncResultTrack]:
        """
        Saves the tags of all tracks in this collection. Use arguments from :py:func:`LocalTrack.save`

        :param tags: Tags to be updated.
        :param replace: Destructively replace tags in each file.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: A map of the :py:class:`LocalTrack` saved to its result as a :py:class:`SyncResultTrack` object
            only for tracks that were saved or would have been saved in the case of a dry run.
        """
        async def _save_track(track: T) -> tuple[T, SyncResultTrack]:
            return track, await track.save(tags=tags, replace=replace, dry_run=dry_run)

        # WARNING: making this run asynchronously will break tqdm; bar will get stuck after 1-2 ticks
        bar = self.logger.get_synchronous_iterator(
            self.tracks, desc="Updating tracks", unit="tracks"
        )
        results = dict([await _save_track(track) for track in bar])
        return {track: result for track, result in results.items() if result.saved or result.updated}

    def log_save_tracks_result(self, results: Mapping[T, SyncResultTrack]) -> None:
        """Log stats from the results of a ``save_tracks`` operation"""
        if not results:
            return

        max_width = get_max_width([track.path for track in results], max_width=80)

        self.logger.stat("\33[1;96mSaved tags to the following tracks: \33[0m")
        for track, result in results.items():
            saved = "\33[92mSAVED" if result.saved else "\33[91mNOT SAVED"
            tags = ', '.join(tag.name for tag in result.updated.keys())
            if not tags:
                tags = "No tags updated"
            self.logger.stat(
                f"\33[97m{align_string(track.path, max_width=max_width, truncate_left=True)} \33[0m| "
                f"{saved} \33[0m| \33[94m{tags} \33[0m"
            )

    def merge_tracks(self, tracks: Collection[Track], tags: UnitIterable[TagField] = Fields.ALL) -> None:
        """
        Merge this collection with another collection or list of items
        by performing an inner join on a given set of tags

        :param tracks: List of items or :py:class:`MusifyCollection` to merge with
        :param tags: List of tags to merge on.
        """
        # noinspection PyTypeChecker
        tag_names = set(TagField.__tags__) if tags == Fields.ALL else set(TagField.to_tags(tags))
        tag_order = [tag for field in TagFields.all(only_tags=True) for tag in field.to_tag()]
        tag_names = sorted(tag_names, key=lambda x: tag_order.index(x))

        if isinstance(self, Library | LocalCollection):  # log status message and use progress bar for libraries
            self.logger.info(
                f"\33[1;95m  >\33[1;97m "
                f"Merging library of {len(self)} items with {len(tracks)} items on tags: "
                f"{', '.join(tag_names)} \33[0m"
            )
            tracks = self.logger.get_synchronous_iterator(tracks, desc="Merging library", unit="tracks")

        tags = to_collection(tags)
        if Fields.IMAGES in tags or Fields.ALL in tags:
            tag_names.append("image_links")

        for track in tracks:  # perform the merge
            track_in_collection: T = next((t for t in self.tracks if t == track), None)
            if not track_in_collection:  # skip if the item does not exist in this collection
                continue

            for tag in tag_names:  # merge on each tag
                if hasattr(track, tag):
                    track_in_collection[tag] = track[tag]

        if isinstance(self, Library | LocalCollection):
            self.logger.print_line()


class LocalCollectionFiltered[T: LocalItem](LocalCollection[T], metaclass=ABCMeta):
    """
    Generic class for storing and filtering on a collection of local tracks
    with methods for enriching the attributes of this object from the attributes of the collection of tracks

    :param tracks: A list of loaded tracks.
    :param name: The name of this collection.
        If given, the object only stores tracks that match the name given on the attribute of this object.
        If None, the list of tracks given are taken to be all the tracks contained in this collection.
    :raise LocalCollectionError: If the given tracks contain more than one unique value
        for the attribute of this collection when name is None.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
    """

    __slots__ = ("_tag_key", "_name", "_tracks")

    @property
    def name(self):
        """The name of the key property of this collection"""
        return self._name

    @property
    def tracks(self):
        return self._tracks

    @property
    def artists(self) -> list[str]:
        """List of artists ordered by frequency of appearance on the tracks in this collection"""
        return get_most_common_values(track.artist for track in self.tracks if track.artist)

    def __init__(
            self,
            tracks: Collection[T],
            name: str | None = None,
            remote_wrangler: RemoteDataWrangler = None
    ):
        super().__init__(remote_wrangler=remote_wrangler)
        if len(tracks) == 0:
            raise LocalCollectionError("No tracks were given")

        # get the tag key dynamically from the name of this class
        self._tag_key = self._pascal_to_snake(self.__class__.__name__.removeprefix("Local"))

        if name is None:  # attempt to determine the name of this collection from the given tracks
            names: list[UnitCollection[str]] = [track[self._tag_key] for track in tracks if track[self._tag_key]]
            names: set[str] = {(n for n in name) if isinstance(name, (list, set, tuple)) else name for name in names}

            if len(names) == 0:  # no valid name found
                raise LocalCollectionError(f"No {self._tag_key.rstrip('s')}s found in the given tracks")
            elif len(names) != 1:  # too many names found
                raise LocalCollectionError(
                    f"Too many {self._tag_key.rstrip('s')}s found in the given tracks. "
                    f"Only provide tracks from the same {self._tag_key.rstrip('s')}."
                )

            self._name: str = names.pop()
            self._tracks: list[T] = to_collection(tracks, list)
        else:  # match tracks with a tag equal to the given name for this collection
            self._name: str = name
            self._tracks: list[T] = self._get_matching_tracks(tracks)

    def _get_matching_tracks(self, tracks: Iterable[T]) -> list[T]:
        """Get a list of tracks that match this collection's name"""
        matched: list[T] = []
        for track in tracks:
            value = track[self._tag_key]
            if isinstance(value, str) and self.name == value:
                matched.append(track)
            elif isinstance(value, Container) and self.name in value:
                matched.append(track)

        return matched


class LocalFolder(LocalCollectionFiltered[LocalTrack], Folder[LocalTrack]):
    """
    Object representing a collection of tracks in a folder on the local drive.

    :param tracks: A list of loaded tracks.
    :param name: The name of this folder.
        If given, the object only stores tracks that match the folder ``name`` given.
        If None, the list of tracks given are taken to be all the tracks contained in this folder.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
    :raise LocalCollectionError: If the given tracks contain more than one unique value for
        ``folder`` when name is None.
    """

    __slots__ = ()
    __attributes_classes__ = (Folder, LocalCollection)

    @property
    def albums(self):
        return get_most_common_values(track.album for track in self.tracks if track.album)

    @property
    def genres(self) -> list[str]:
        """List of genres ordered by frequency of appearance on the tracks in this collection"""
        genres = (genre for track in self.tracks for genre in (track.genres if track.genres else []))
        return get_most_common_values(genres)

    @property
    def compilation(self):
        """Folder is a compilation if over 50% of tracks are marked as compilation"""
        return (sum(track.compilation is True for track in self.tracks) / len(self.tracks)) > 0.5

    def __init__(
            self,
            tracks: Collection[LocalTrack] = (),
            name: str | None = None,
            remote_wrangler: RemoteDataWrangler = None
    ):
        super().__init__(tracks=tracks, name=name, remote_wrangler=remote_wrangler)
        self.tracks.sort(key=lambda x: x.filename or _max_str)

    @classmethod
    async def load_folder(cls, path: str | Path | None, remote_wrangler: RemoteDataWrangler = None) -> Self:
        """
        Load tracks in a folder at the given ``path``.

        :param path: The path of the folder to load.
        :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object
            for processing URIs on tracks. If given, the wrangler can be used when calling __get_item__
            to get an item from the collection from its URI.
        """
        if not path.is_dir():
            raise UnexpectedPathError(path, "Path must be a directory")

        # load tracks in the folder
        tasks = asyncio.gather(*[load_track(p) for p in path.glob("*") if p.suffix in TRACK_FILETYPES])
        return cls(tracks=await tasks, name=path.name, remote_wrangler=remote_wrangler)


class LocalAlbum(LocalCollectionFiltered[LocalTrack], Album[LocalTrack]):
    """
    Object representing a collection of tracks of an album.

    :param tracks: A list of loaded tracks.
    :param name: The name of this album.
        If given, the object only stores tracks that match the album ``name`` given.
        If None, the list of tracks given are taken to be all the tracks for this album.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
    :raise LocalCollectionError: If the given tracks contain more than one unique value for
        ``album`` when name is None.
    """

    __slots__ = ("_image_links",)
    __attributes_classes__ = (Album, LocalCollection)

    @property
    def album_artist(self):
        """The most common artist on this album"""
        artists = get_most_common_values(artist for track in self.tracks if track.artist for artist in track.artists)
        return artists[0] if artists else None

    @property
    def genres(self) -> list[str]:
        """List of genres ordered by frequency of appearance on the tracks in this collection"""
        genres = (genre for track in self.tracks for genre in (track.genres if track.genres else []))
        return get_most_common_values(genres)

    @property
    def date(self):
        """
        A :py:class:`date` object representing the release date of this album.
        Determined by the most common release date of all tracks on this album.
        """
        values = get_most_common_values(track.date for track in self.tracks if track.date)
        return values[0] if values else None

    @property
    def year(self):
        """The most common release year of all tracks on this album"""
        values = get_most_common_values(track.year for track in self.tracks if track.year)
        return values[0] if values else None

    @property
    def month(self):
        """The most common release month of all tracks on this album"""
        values = get_most_common_values(
            (track.year, track.month) for track in self.tracks if track.year and track.month
        )
        return values[0][1] if values else None

    @property
    def day(self):
        """The most common release day of all tracks on this album"""
        values = get_most_common_values(
            (track.year, track.month, track.day) for track in self.tracks if track.year and track.month and track.day
        )
        return values[0][2] if values else None

    @property
    def compilation(self):
        """Album is a compilation if over 50% of tracks are marked as compilation"""
        return (sum(track.compilation is True for track in self.tracks) / len(self.tracks)) > 0.5

    @property
    def image_links(self):
        return self._image_links

    @property
    def has_image(self):
        return any(track.has_image for track in self.tracks)

    @property
    def rating(self):
        """Average rating of all tracks on this album"""
        ratings = tuple(track.rating for track in self.tracks if track.rating is not None)
        return sum(ratings) / len(ratings) if len(ratings) > 0 else None

    def __init__(
            self, tracks: Collection[LocalTrack], name: str | None = None, remote_wrangler: RemoteDataWrangler = None
    ):
        super().__init__(tracks=tracks, name=name, remote_wrangler=remote_wrangler)
        self.tracks.sort(
            key=lambda x: (x.disc_number or sys.maxsize, x.track_number or sys.maxsize, x.filename or _max_str)
        )
        self._image_links: dict[str, str] = {}


class LocalArtist(LocalCollectionFiltered[LocalTrack], Artist[LocalTrack]):
    """
    Object representing a collection of tracks by a single artist.

    :param tracks: A list of loaded tracks.
    :param name: The name of this artist.
        If given, the object only stores tracks that match the artist ``name`` given.
        If None, the list of tracks given are taken to be all the tracks by this artist.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
    :raise LocalCollectionError: If the given tracks contain more than one unique value for
        ``artist`` when name is None.
    """

    __slots__ = ()
    __attributes_classes__ = (Artist, LocalCollection)

    @property
    def albums(self):
        return get_most_common_values(track.album for track in self.tracks if track.album)

    @property
    def genres(self) -> list[str]:
        """List of genres ordered by frequency of appearance on the tracks in this collection"""
        genres = (genre for track in self.tracks for genre in (track.genres if track.genres else []))
        return get_most_common_values(genres)

    @property
    def rating(self):
        """Average rating of all tracks by this artist"""
        ratings = tuple(track.rating for track in self.tracks if track.rating is not None)
        return sum(ratings) / len(ratings) if len(ratings) > 0 else None

    def __init__(
            self, tracks: Collection[LocalTrack], name: str | None = None, remote_wrangler: RemoteDataWrangler = None
    ):
        super().__init__(tracks=tracks, name=name, remote_wrangler=remote_wrangler)
        self.tracks.sort(
            key=lambda x: (x.album or _max_str,
                           x.disc_number or sys.maxsize,
                           x.track_number or sys.maxsize,
                           x.filename or _max_str)
        )


class LocalGenres(LocalCollectionFiltered[LocalTrack], Genre[LocalTrack]):
    """
    Object representing a collection of tracks within a genre.

    :param tracks: A list of loaded tracks.
    :param name: The name of this genre.
        If given, the object only stores tracks that match the genre ``name`` given.
        If None, the list of tracks given are taken to be all the tracks within this genre.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
    :raise LocalCollectionError: If the given tracks contain more than one unique value for
        ``genre`` when name is None.
    """

    __slots__ = ()
    __attributes_classes__ = (Genre, LocalCollection)

    @property
    def albums(self):
        return get_most_common_values(track.album for track in self.tracks if track.album)

    @property
    def related_genres(self) -> list[str]:
        genres = (genre for track in self.tracks for genre in (track.genres if track.genres else []))
        return get_most_common_values(genres)

    def __init__(
            self, tracks: Collection[LocalTrack], name: str | None = None, remote_wrangler: RemoteDataWrangler = None
    ):
        super().__init__(tracks=tracks, name=name, remote_wrangler=remote_wrangler)
        self.tracks.sort(
            key=lambda x: (x.artist or _max_str,
                           x.album or _max_str,
                           x.disc_number or sys.maxsize,
                           x.track_number or sys.maxsize,
                           x.filename or _max_str)
        )



"""
All remote classes and operations with remote objects, files, tracks, albums, playlists, libraries etc.
"""
from musify.libraries.remote.spotify import SOURCE_NAME as SPOTIFY

REMOTE_SOURCES = frozenset([SPOTIFY])



"""
The framework for interacting with a remote API.

All methods that interact with the API should return raw, unprocessed responses.
"""
import logging
from abc import ABCMeta, abstractmethod
from collections.abc import Collection, MutableMapping, Mapping, Sequence, Iterable
from typing import Any, Self

from aiorequestful.auth import Authoriser
from aiorequestful.cache.backend.base import ResponseCache
from aiorequestful.cache.exception import CacheError
from aiorequestful.cache.session import CachedSession
from aiorequestful.request import RequestHandler
from aiorequestful.response.payload import JSONPayloadHandler
from aiorequestful.types import ImmutableJSON, JSON
from yarl import URL

from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.types import APIInputValueSingle, APIInputValueMulti, RemoteIDType, RemoteObjectType
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.logger import MusifyLogger
from musify.types import UnitSequence, UnitList
from musify.utils import align_string, to_collection


class RemoteAPI[A: Authoriser](metaclass=ABCMeta):
    """
    Collection of endpoints for a remote API.
    See :py:class:`RequestHandler` and :py:class:`Authoriser`
    for more info on which params to pass to authorise and execute requests.

    :param wrangler: The :py:class:`RemoteDataWrangler` for this API type.
    :param authoriser: The authoriser to use when authorising requests to the API.
    :param cache: When given, attempt to use this cache for certain request types before calling the API.
        Repository and cachable request types can be set up by child classes.
    """

    __slots__ = ("logger", "handler", "wrangler", "user_data", "user_playlist_data")

    #: Map of :py:class:`RemoteObjectType` for remote collections
    #: to the  :py:class:`RemoteObjectType` of the items they hold
    collection_item_map = {
        RemoteObjectType.PLAYLIST: RemoteObjectType.TRACK,
        RemoteObjectType.ALBUM: RemoteObjectType.TRACK,
        RemoteObjectType.AUDIOBOOK: RemoteObjectType.CHAPTER,
        RemoteObjectType.SHOW: RemoteObjectType.EPISODE,
    }
    #: A set of possible saved item types that can be retrieved for the currently authenticated user
    user_item_types = (
            set(collection_item_map) | {RemoteObjectType.TRACK, RemoteObjectType.ARTIST, RemoteObjectType.EPISODE}
    )
    #: The key to use when getting an ID from a response
    id_key = "id"
    #: The key to use when getting the URL from a response
    url_key = "href"

    @property
    @abstractmethod
    def user_id(self) -> str | None:
        """ID of the currently authorised user"""
        raise NotImplementedError

    @property
    @abstractmethod
    def user_name(self) -> str | None:
        """Name of the currently authorised user"""
        raise NotImplementedError

    @property
    def url(self) -> URL:
        """The base URL of the API"""
        return self.wrangler.url_api

    @property
    def source(self) -> str:
        """The name of the API service"""
        return self.wrangler.source

    def __init__(self, authoriser: A, wrangler: RemoteDataWrangler, cache: ResponseCache | None = None):
        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

        #: A :py:class:`RemoteDataWrangler` object for processing URIs
        self.wrangler = wrangler

        #: The :py:class:`RequestHandler` for handling authorised requests to the API
        self.handler: RequestHandler[A, JSON] = RequestHandler.create(
            authoriser=authoriser, cache=cache, payload_handler=JSONPayloadHandler(indent=2),
        )

        #: Stores the loaded user data for the currently authorised user
        self.user_data: dict[str, Any] = {}
        #: Stores the loaded user playlists data for the currently authorised user
        self.user_playlist_data: dict[str, dict[str, Any]] = {}

    async def __aenter__(self) -> Self:
        await self.handler.__aenter__()

        try:
            await self._setup_cache()
        except CacheError:
            pass

        session = self.handler.session
        if isinstance(session, CachedSession):
            for repository in session.cache.values():
                # all repositories must use the same payload handler as the request handler
                # for it to function correctly
                repository.settings.payload_handler = self.handler.payload_handler

        await self.load_user()
        await self.load_user_playlists()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.handler.__aexit__(exc_type, exc_val, exc_tb)

    @abstractmethod
    async def _setup_cache(self) -> None:
        """Set up the repositories and repository getter on the self.handler.session's cache."""
        raise NotImplementedError

    async def authorise(self) -> Self:
        """
        Main method for authorisation, tests/refreshes/reauthorises as needed

        :return: Self.
        :raise APIError: If the token cannot be validated.
        """
        await self.handler.authorise()
        return self

    async def close(self) -> None:
        """Close the current session. No more requests will be possible once this has been called."""
        await self.handler.close()

    ###########################################################################
    ## Misc helpers
    ###########################################################################
    async def load_user(self) -> None:
        """Load and store user data for the currently authorised user in this API object"""
        self.user_data = await self.get_self()

    @abstractmethod
    async def load_user_playlists(self) -> None:
        """Load and store user playlists data for the currently authorised user in this API object"""
        raise NotImplementedError

    @staticmethod
    def _merge_results_to_input_mapping(
            original: MutableMapping[str, Any], response: Mapping[str, Any], clear: bool = True,
    ) -> None:
        if clear:
            original.clear()
        original |= response

    def _merge_results_to_input(
            self,
            original: UnitSequence[ImmutableJSON] | UnitSequence[RemoteResponse],
            responses: UnitList[ImmutableJSON],
            ordered: bool = True,
            clear: bool = True,
    ) -> None:
        """
        If API response type given on input, update with new results.
        Assumes on a one-to-one relationship between ``original`` and the list of ``results``.

        :param original: The original values given to the function.
        :param responses: The new results from the API.
        :param ordered: When True, function assumes the order of items in ``original`` and ``results`` is the same.
            When False, the function will attempt to match each input value to each result by matching on
            the ``id`` key of each dictionary.
        :param ordered: When True, clear the original value before merging, completely replacing all original data.
        """
        if isinstance(original, str) or not isinstance(original, Collection | RemoteResponse):
            return

        if isinstance(original, RemoteResponse):
            original = [original]
        elif not isinstance(original, Sequence):
            original = to_collection(original)
        if not isinstance(responses, Sequence):
            responses = to_collection(responses, list)

        original = [item.response if isinstance(item, RemoteResponse) else item for item in original]

        valid_types_input = all(isinstance(item, MutableMapping) for item in original)
        valid_types_responses = all(isinstance(item, MutableMapping) for item in responses)
        valid_lengths = len(original) == len(responses)
        if not all((valid_types_input, valid_types_responses, valid_lengths)):
            self.logger.debug(
                "Could not merge responses to given user input | "
                "Reason: assertions failed | "
                f"{valid_types_input=} {valid_types_responses=} {valid_lengths=} | "
            )
            return

        if not ordered:
            expected_keys_input = all(self.id_key in item for item in original)
            expected_keys_responses = all(self.id_key in item for item in responses)
            if not all((expected_keys_input, expected_keys_responses)):
                self.logger.debug(
                    "Could not merge responses to given user input | "
                    f"Reason: unordered and cannot order on {self.id_key=} | "
                    f"{expected_keys_input=} {expected_keys_responses=}"
                )
                return

            id_ordered = [response[self.id_key] for response in original]
            responses.sort(key=lambda response: id_ordered.index(response[self.id_key]))

        for item, response in zip(original, responses):
            self._merge_results_to_input_mapping(original=item, response=response, clear=clear)

    @staticmethod
    def _refresh_responses(responses: Any, skip_checks: bool = False) -> None:
        if isinstance(responses, RemoteResponse):
            responses = [responses]
        if not isinstance(responses, Iterable) or not any(isinstance(r, RemoteResponse) for r in responses):
            return

        for response in responses:
            if isinstance(response, RemoteResponse):
                response.refresh(skip_checks=skip_checks)

    def print_item(
            self, i: int, name: str, uri: str, length: float = 0, total: int = 1, max_width: int = 50
    ) -> None:
        """
        Pretty print item data for displaying to the user.
        Format = ``<i> - <name> | <length> | <URI> - <URL>``.

        :param i: The position of this item in the collection.
        :param name: The name of the item.
        :param uri: The URI of the item.
        :param length: The duration of the item in seconds.
        :param total: The total number of items in the collection
        :param max_width: The maximum width to print names as. Any name lengths longer than this will be truncated.
        """
        self.logger.print_message(
            f"\33[92m{str(i).zfill(len(str(total)))} \33[0m- "
            f"\33[97m{align_string(name, max_width=max_width)} \33[0m| "
            f"\33[91m{str(int(length // 60)).zfill(2)}:{str(round(length % 60)).zfill(2)} \33[0m| "
            f"\33[93m{uri} \33[0m- "
            f"{self.wrangler.convert(uri, type_in=RemoteIDType.URI, type_out=RemoteIDType.URL_EXT)}"
        )

    @abstractmethod
    async def print_collection(
            self,
            value: APIInputValueSingle[RemoteResponse] | None = None,
            kind: RemoteIDType | None = None,
            limit: int = 20,
    ) -> None:
        """
        Pretty print collection data for displaying to the user.
        Runs :py:meth:`print_item()` for each item in the collection.

        ``value`` may be:
            * A string representing a URL/URI/ID.
            * A remote API JSON response for a collection with a valid ID value under an ``id`` key.
            * A RemoteResponse representing some remote collection of items.

        :param value: The value representing some remote collection. See description for allowed value types.
        :param kind: When an ID is provided, give the kind of ID this is here.
            If None and ID is given, user will be prompted to give the kind anyway.
        :param limit: The number of results to call per request and,
            therefore, the number of items in each printed block.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_playlist_url(self, playlist: APIInputValueSingle[RemoteResponse]) -> URL:
        """
        Determine the type of the given ``playlist`` and return its API URL.
        If type cannot be determined, attempt to find the playlist in the
        list of the currently authenticated user's playlists.

        :param playlist: In URL/URI/ID form, or the name of one of the currently authenticated user's playlists.
        :return: The playlist URL.
        :raise RemoteIDTypeError: Raised when the function cannot determine the item type of
            the input ``playlist``. Or when it does not recognise the type of the input ``playlist`` parameter.
        """
        raise NotImplementedError

    ###########################################################################
    ## Core - GET endpoints
    ###########################################################################
    @abstractmethod
    async def get_self(self, update_user_data: bool = True) -> dict[str, Any]:
        """
        ``GET`` - Get API response for information on current user

        :param update_user_data: When True, update the ``_user_data`` stored in this API object."""
        raise NotImplementedError

    @abstractmethod
    async def query(self, query: str, kind: RemoteObjectType, limit: int = 10) -> list[dict[str, Any]]:
        """
        ``GET`` - Query for items. Modify result types returned with kind parameter

        :param query: Search query.
        :param kind: The remote object type to search for.
        :param limit: Number of results to get and return.
        :return: The response from the endpoint.
        """
        raise NotImplementedError

    ###########################################################################
    ## Item - GET endpoints
    ###########################################################################
    @abstractmethod
    async def extend_items(
            self,
            response: MutableMapping[str, Any] | RemoteResponse,
            kind: RemoteObjectType | str | None = None,
            key: RemoteObjectType | None = None,
            leave_bar: bool | None = None,
    ) -> list[dict[str, Any]]:
        """
        Extend the items for a given API ``response``.
        The function requests each page of the collection returning a list of all items
        found across all pages for this URL.

        Updates the value of the ``items`` key in-place by extending the value of the ``items`` key with new results.

        If a :py:class:`RemoteResponse`, this function will not refresh it with the new response.
        The user must call `refresh` manually after execution.

        :param response: A remote API JSON response for an items type endpoint.
        :param kind: The type of response being extended. Optional, used only for logging.
        :param key: The type of response of the child objects.
        :param leave_bar: When a progress bar is displayed,
            toggle whether this bar should continue to be displayed after the operation is finished.
            When None, allow the logger to decide this setting.
        :return: API JSON responses for each item
        """
        raise NotImplementedError

    @abstractmethod
    async def get_items(
            self,
            values: APIInputValueMulti[RemoteResponse],
            kind: RemoteObjectType | None = None,
            limit: int = 50,
            extend: bool = True,
    ) -> list[dict[str, Any]]:
        """
        ``GET`` - Get information for given ``values``.

        ``values`` may be:
            * A string representing a URL/URI/ID.
            * A MutableSequence of strings representing URLs/URIs/IDs of the same type.
            * A remote API JSON response for a collection.
            * A MutableSequence of remote API JSON responses for a collection.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this updates each response given by merging with the new response.

        If :py:class:`RemoteResponse` values are given, this function will call `refresh` on them.

        :param values: The values representing some remote objects. See description for allowed value types.
            These items must all be of the same type of item i.e. all tracks OR all artists etc.
        :param kind: Item type if given string is ID.
        :param limit: When requests can be batched, size of batches to request.
            This value will be limited to be between ``1`` and ``50``.
        :param extend: When True and the given ``kind`` is a collection of items,
            extend the response to include all items in this collection.
        :return: API JSON responses for each item.
        :raise RemoteObjectTypeError: Raised when the function cannot determine the item type
            of the input ``values``. Or when it does not recognise the type of the input ``values`` parameter.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_tracks(
            self, values: APIInputValueMulti[RemoteResponse], limit: int = 50, *args, **kwargs
    ) -> list[dict[str, Any]]:
        """
        Wrapper for :py:meth:`get_items` which only returns Track type responses.
        See :py:meth:`get_items` for more info.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_user_items(
            self, user: str | None = None, kind: RemoteObjectType = RemoteObjectType.PLAYLIST, limit: int = 50,
    ) -> list[dict[str, Any]]:
        """
        ``GET`` - Get saved items for a given user.

        :param user: The ID of the user to get playlists for. If None, use the currently authenticated user.
        :param kind: Item type to retrieve for the user.
        :param limit: Size of each batch of items to request in the collection items request.
            This value will be limited to be between ``1`` and ``50``.
        :return: API JSON responses for each collection.
        :raise RemoteIDTypeError: Raised when the input ``user`` does not represent a user URL/URI/ID.
        :raise RemoteObjectTypeError: When the given ``kind`` is not a valid user item/collection.
        """
        raise NotImplementedError

    ###########################################################################
    ## Playlist specific endpoints
    ###########################################################################
    @abstractmethod
    async def create_playlist(self, name: str, *args, **kwargs) -> dict[str, Any]:
        """
        ``POST`` - Create an empty playlist for the current user with the given name.

        :param name: Name of playlist to create.
        :return: API JSON response for the created playlist.
        """
        raise NotImplementedError

    @abstractmethod
    async def add_to_playlist(
            self,
            playlist: APIInputValueSingle[RemoteResponse],
            items: Sequence[str],
            limit: int = 50,
            skip_dupes: bool = True
    ) -> int:
        """
        ``POST`` - Add list of tracks to a given playlist.

        :param playlist: One of the following to identify the playlist to clear:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param items: List of URLs/URIs/IDs of the tracks to add.
        :param limit: Size of each batch of IDs to add. This value will be limited to be between ``1`` and ``50``.
        :param skip_dupes: Skip duplicates.
        :return: The number of tracks added to the playlist.
        :raise RemoteIDTypeError: Raised when the input ``playlist`` does not represent
            a playlist URL/URI/ID.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``items``
            are not all tracks or IDs.
        """
        raise NotImplementedError

    async def get_or_create_playlist(self, name: str, *args, **kwargs) -> dict[str, Any]:
        """
        Attempt to find the playlist with the given ``name`` and return it.
        Otherwise, create a new playlist.

        Any given args and kwargs are passed directly onto :py:meth:`create_playlist`
        when a matching loaded playlist is not found.

        When a playlist is created, persist the response back to the loaded playlists in this object.

        :param name: The case-sensitive name of the playlist to get/create.
        :return: API JSON response for the loaded/created playlist.
        """
        if not self.user_playlist_data:
            await self.load_user_playlists()

        response = self.user_playlist_data.get(name)
        if response:
            return response

        return await self.create_playlist(name, *args, **kwargs)

    async def follow_playlist(self, playlist: APIInputValueSingle[RemoteResponse], *args, **kwargs) -> URL:
        """
        ``PUT`` - Follow a given playlist.

        :param playlist: One of the following to identify the playlist to clear:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :return: API URL for playlist.
        """
        raise NotImplementedError

    @abstractmethod
    async def delete_playlist(self, playlist: APIInputValueSingle[RemoteResponse]) -> URL:
        """
        ``DELETE`` - Unfollow/delete a given playlist.
        WARNING: This function will destructively modify your remote playlists.

        :param playlist: One of the following to identify the playlist to clear:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :return: API URL for playlist.
        """
        raise NotImplementedError

    @abstractmethod
    async def clear_from_playlist(
            self,
            playlist: APIInputValueSingle[RemoteResponse],
            items: Collection[str] | None = None,
            limit: int = 100
    ) -> int:
        """
        ``DELETE`` - Clear tracks from a given playlist.
        WARNING: This function can destructively modify your remote playlists.

        :param playlist: One of the following to identify the playlist to clear:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param items: List of URLs/URIs/IDs of the tracks to remove. If None, clear all songs from the playlist.
        :param limit: Size of each batch of IDs to clear in a single request.
            This value will be limited to be between ``1`` and ``100``.
        :return: The number of tracks cleared from the playlist.
        :raise RemoteIDTypeError: Raised when the input ``playlist`` does not represent a playlist URL/URI/ID.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``items``
            are not all tracks or IDs.
        """
        raise NotImplementedError



"""
Functionality relating to reading and writing images.
"""
from http.client import HTTPResponse
from io import BytesIO
from pathlib import Path
from urllib.error import URLError
from urllib.request import urlopen, Request

from PIL import Image, UnidentifiedImageError
from yarl import URL

from musify.file.exception import ImageLoadError


def open_image(source: str | bytes | Path | URL | Request) -> Image.Image:
    """
    Open Image object from a given URL or file path

    :return: The loaded :py:class:`Image.Image`
    :raise ImageLoadError: If the image cannot be loaded.
    :raise ModuleImportError: If required modules are not installed.
    """

    try:  # open image from link
        if isinstance(source, URL):
            source = str(source)

        if isinstance(source, Request) or (isinstance(source, str) and source.startswith("http")):
            response: HTTPResponse = urlopen(source)
            image = Image.open(response)
            response.close()
            return image

        elif not isinstance(source, Request):
            return Image.open(source)

    except (URLError, FileNotFoundError, UnidentifiedImageError):
        pass
    raise ImageLoadError(source, "Failed to open image")


def get_image_bytes(image: Image.Image) -> bytes:
    """
    Extracts bytes from a given Image file.

    :raise ModuleImportError: If required modules are not installed.
    """
    image_bytes_arr = BytesIO()
    image.save(image_bytes_arr, format=image.format)
    return image_bytes_arr.getvalue()
